# -*- coding: utf-8 -*-


"""
Routines for converting error responses to appropriate exceptions.
"""
from . import exceptions

__all__ = ['raise_if_error']

# Patterns used to map exception strings to classes.

# First, exceptions for which the messages start with the error name,
# and then contain the error description wrapped in [].
exceptions_by_name = dict((name, getattr(exceptions, name))
for name in (
    "DocumentAlreadyExistsEngineException",
    "DocumentAlreadyExistsException",
    "TypeMissingException",
    "VersionConflictEngineException",
    'ClusterBlockException',
    'ElasticSearchIllegalArgumentException',
    'IndexAlreadyExistsException',
    'IndexMissingException',
    'InvalidIndexNameException',
    'MapperParsingException',
    'ReduceSearchPhaseException',
    'ReplicationShardOperationFailedException',
    'SearchPhaseExecutionException',
    'DocumentMissingException',
    )
)

# Second, patterns for exceptions where the message is just the error
# description, and doesn't contain an error name.  These patterns are matched
# at the end of the exception.
exception_patterns_trailing = {
    '] missing': exceptions.NotFoundException,
    '] Already exists': exceptions.AlreadyExistsException,
    }

def raise_if_error(status, result, request=None):
    """Raise an appropriate exception if the result is an error.

    Any result with a status code of 400 or higher is considered an error.

    The exception raised will either be an ElasticSearchException, or a more
    specific subclass of ElasticSearchException if the type is recognised.

    The status code and result can be retrieved from the exception by accessing its
    status and result properties.

    Optionally, this can take the original RestRequest instance which generated
    this error, which will then get included in the exception.

    """
    assert isinstance(status, int)

    if status < 400:
        return

    if status == 404 and isinstance(result, dict) and 'error' not in result:
        raise exceptions.NotFoundException("Item not found", status, result, request)

    if not isinstance(result, dict) or 'error' not in result:
        raise exceptions.ElasticSearchException("Unknown exception type: %d, %s" % (status, result), status,
            result, request)

    error = result['error']
    if '; nested: ' in error:
        error_list = error.split('; nested: ')
        error = error_list[len(error_list) - 1]

    bits = error.split('[', 1)
    if len(bits) == 2:
        excClass = exceptions_by_name.get(bits[0], None)
        if excClass is not None:
            msg = bits[1]
            if msg.endswith(']'):
                msg = msg[:-1]
            '''
            if request:
                msg += ' (' + str(request) + ')'
            '''
            raise excClass(msg, status, result, request)

    for pattern, excClass in list(exception_patterns_trailing.items()):
        if not error.endswith(pattern):
            continue
            # For these exceptions, the returned value is the whole descriptive
        # message.
        raise excClass(error, status, result, request)

    raise exceptions.ElasticSearchException(error, status, result, request)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import base64

try:
    # python3
    from urllib.parse import quote as _quote
except ImportError:
    # Python 2
    from urllib import quote as _quote
import uuid
import six

__all__ = ['clean_string', "ESRange", "ESRangeOp", "string_b64encode", "string_b64decode", "make_path", "make_id"]


def quote(value):
    value = value.encode('utf8', 'ignore') if isinstance(value, six.string_types) else str(value)
    return _quote(value, safe='')


def plain_concepts_extractor(explanation):
    if "description" in explanation and explanation["description"].startswith(u"weight("):
        description = explanation["description"]
        return [(description.split(u"weight(")[1].split(u")")[0].rsplit(u" in ", 1)[0], explanation["value"])]
    results = []
    for detail in explanation.get("details", []):
        results.extend(plain_concepts_extractor(detail))
    return results


def make_id(value):
    """
    Build a string id from a value
    :param value: a text value
    :return: a string
    """
    if isinstance(value, six.string_types):
        value = value.encode("utf8", "ignore")
    from hashlib import md5

    val = uuid.UUID(bytes=md5(value).digest(), version=4)

    return base64.urlsafe_b64encode(val.get_bytes())[:-2]


def make_path(*path_components):
    """
    Smash together the path components. Empty components will be ignored.
    """
    path_components = [quote(component) for component in path_components if component]
    path = '/'.join(path_components)
    if not path.startswith('/'):
        path = '/' + path
    return path


def string_b64encode(s):
    """
    This function is useful to convert a string to a valid id to be used in ES.
    You can use it to generate an ID for urls or some texts
    """
    return base64.urlsafe_b64encode(s).strip('=')


def string_b64decode(s):
    return base64.urlsafe_b64decode(s + '=' * (len(s) % 4))

# Characters that are part of Lucene query syntax must be stripped
# from user input: + - && || ! ( ) { } [ ] ^ " ~ * ? : \
# See: http://lucene.apache.org/java/3_0_2/queryparsersyntax.html#Escaping
SPECIAL_CHARS = [33, 34, 38, 40, 41, 42, 45, 58, 63, 91, 92, 93, 94, 123, 124, 125, 126]
UNI_SPECIAL_CHARS = dict((c, None) for c in SPECIAL_CHARS)
STR_SPECIAL_CHARS = ''.join([chr(c) for c in SPECIAL_CHARS])


class EqualityComparableUsingAttributeDictionary(object):
    """
    Instances of classes inheriting from this class can be compared
    using their attribute dictionary (__dict__). See GitHub issue
    128 and http://stackoverflow.com/q/390640
    """

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __ne__(self, other):
        return not self == other


class ESRange(EqualityComparableUsingAttributeDictionary):

    def __init__(self, field, from_value=None, to_value=None, include_lower=None,
                 include_upper=None, **kwargs):
        self.field = field
        self.from_value = from_value
        self.to_value = to_value
        self.include_lower = include_lower
        self.include_upper = include_upper

    def negate(self):
        """Reverse the range"""
        self.from_value, self.to_value = self.to_value, self.from_value
        self.include_lower, self.include_upper = self.include_upper, self.include_lower

    def serialize(self):
        filters = {}
        if self.from_value is not None:
            include_lower = "gte" if self.include_lower else "gt"
            filters[include_lower] = self.from_value

        if self.to_value is not None:
            include_upper = "lte" if self.include_upper else "lt"
            filters[include_upper] = self.to_value

        return self.field, filters


class ESRangeOp(ESRange):
    def __init__(self, field, op1, value1, op2=None, value2=None):
        from_value = to_value = include_lower = include_upper = None
        for op, value in ((op1, value1), (op2, value2)):
            if op == "gt":
                from_value = value
                include_lower = False
            elif op == "gte":
                from_value = value
                include_lower = True
            if op == "lt":
                to_value = value
                include_upper = False
            elif op == "lte":
                to_value = value
                include_upper = True
        super(ESRangeOp, self).__init__(field, from_value, to_value,
                                        include_lower, include_upper)


class TermsLookup(EqualityComparableUsingAttributeDictionary):

    def __init__(self, index, type, id, path, routing=None, cache=None):

        self.index = index
        self.type = type
        self.id = id
        self.path = path
        self.routing = routing
        self.cache = cache

    def serialize(self):
        data = {"index": self.index, "type": self.type,
                "id": self.id, "path": self.path}
        if self.routing:
            data['routing'] = self.routing
        if self.cache:
            data['cache'] = self.cache
        return data


def clean_string(text):
    """
    Remove Lucene reserved characters from query string
    """
    if isinstance(text, six.string_types):
        return text.translate(UNI_SPECIAL_CHARS).strip()
    return text.translate(None, STR_SPECIAL_CHARS).strip()


def keys_to_string(data):
    """
    Function to convert all the unicode keys in string keys
    """
    if isinstance(data, dict):
        for key in list(data.keys()):
            if isinstance(key, six.string_types):
                value = data[key]
                val = keys_to_string(value)
                del data[key]
                data[key.encode("utf8", "ignore")] = val
    return data

def get_unicode_string(string):
  try:
    return str(string, 'utf8')
  except TypeError: # python 2.x compatible
    return unicode(string)




#!/usr/bin/env python
#
# Copyright (C) 2010 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.

"""This is a template for constructing an external archiver for situations
where one wants to archive posts in Mailman's pipermail archive, but also
wants to invoke some other process on the archived message after its URL
and/or path are known.

It assumes this is invoked by mm_cfg.py settings like
PUBLIC_EXTERNAL_ARCHIVER = '/path/to/Ext_Arch.py %(hostname)s %(listname)s'
PRIVATE_EXTERNAL_ARCHIVER = '/path/to/Ext_Arch.py %(hostname)s %(listname)s'

The path in the sys.path.insert() below must be adjusted to the actual path
to Mailman's bin/ directory, or you can simply put this script in Mailman's
bin/ directory and it will work without the sys.path.insert() and of course,
you must add the code you want to the ext_process function.
"""

import sys
sys.path.insert(0, '/usr/local/mailman/bin') # path to your mailman dir
import paths

import os
import email
import time

from cStringIO import StringIO

from Mailman import Message
from Mailman import MailList
from Mailman.Archiver import HyperArch
from Mailman.Logging.Syslog import syslog
from Mailman.Logging.Utils import LogStdErr

# For debugging, log stderr to Mailman's 'debug' log
LogStdErr('debug', 'mailmanctl', manual_reprime=0)

def ext_process(listname, hostname, url, filepath, msg):
    """Here's where you put your code to deal with the just archived message.

    Arguments here are the list name, the host name, the URL to the just
    archived message, the file system path to the just archived message and
    the message object.

    These can be replaced or augmented as needed.
    """
    from pyes import ES
    from pyes.exceptions import ClusterBlockException, NoServerAvailable
    import datetime

    #CHANGE this settings to reflect your configuration
    _ES_SERVERS = ['127.0.0.1:9500'] # I prefer thrift
    _indexname = "mailman"
    _doctype = "mail"
    date = datetime.datetime.today()

    try:
        iconn = ES(_ES_SERVERS)
        status = None
        try:
            status = iconn.status(_indexname)
            logger.debug("Indexer status:%s" % status)
        except:
            iconn.create_index(_indexname)
            time.sleep(1)
            status = iconn.status(_indexname)
            mappings = { u'text': {'boost': 1.0,
                                     'index': 'analyzed',
                                     'store': 'yes',
                                     'type': u'string',
                                     "term_vector" : "with_positions_offsets"},
                             u'url': {'boost': 1.0,
                                        'index': 'not_analyzed',
                                        'store': 'yes',
                                        'type': u'string',
                                        "term_vector" : "no"},
                             u'title': {'boost': 1.0,
                                        'index': 'analyzed',
                                        'store': 'yes',
                                        'type': u'string',
                                        "term_vector" : "with_positions_offsets"},
                             u'date': {'store': 'yes',
                                        'type': u'date'}}
            time.sleep(1)
            status = iconn.put_mapping(_doctype, mappings, _indexname)


        data = dict(url=url,
                    title=msg.get('subject'),
                    date=date,
                    text=str(msg)
                    )
        iconn.index(data, _indexname, _doctype)

        syslog('debug', 'listname: %s, hostname: %s, url: %s, path: %s, msg: %s',
               listname, hostname, url, filepath, msg)
    except ClusterBlockException:
        syslog('error', 'Cluster in revocery state: listname: %s, hostname: %s, url: %s, path: %s, msg: %s',
               listname, hostname, url, filepath, msg)
    except NoServerAvailable:
        syslog('error', 'No server available: listname: %s, hostname: %s, url: %s, path: %s, msg: %s',
               listname, hostname, url, filepath, msg)
    except:
        import traceback
        syslog('error', 'Unknown: listname: %s, hostname: %s, url: %s, path: %s, msg: %s\nstacktrace: %s',
               listname, hostname, url, filepath, msg, repr(traceback.format_exc()))

    return

def main():
    """This is the mainline.

    It first invokes the pipermail archiver to add the message to the archive,
    then calls the function above to do whatever with the archived message
    after it's URL and path are known.
    """

    listname = sys.argv[2]
    hostname = sys.argv[1]

    # We must get the list unlocked here because it is already locked in
    # ArchRunner. This is safe because we aren't actually changing our list
    # object. ArchRunner's lock plus pipermail's archive lock will prevent
    # any race conditions.
    mlist = MailList.MailList(listname, lock=False)

    # We need a seekable file for processUnixMailbox()
    f = StringIO(sys.stdin.read())

    # If we don't need a Message.Message instance, we can skip the next and
    # the imports of email and Message above.
    msg = email.message_from_file(f, Message.Message)

    h = HyperArch.HyperArchive(mlist)
    # Get the message number for the next message
    sequence = h.sequence
    # and add the message.
    h.processUnixMailbox(f)
    f.close()

    # Get the archive name, etc.
    archive = h.archive
    msgno = '%06d' % sequence
    filename = msgno + '.html'
    filepath = os.path.join(h.basedir, archive, filename)
    h.close()

    url = '%s%s/%s' % (mlist.GetBaseArchiveURL(), archive, filename)

    ext_process(listname, hostname, url, filepath, msg)

if __name__ == '__main__':
    main()



from .utils import EqualityComparableUsingAttributeDictionary
from .filters import Filter, TermFilter, TermsFilter, ANDFilter, NotFilter


class AggFactory(EqualityComparableUsingAttributeDictionary):

    def __init__(self):
        self.aggs = []

    def add(self, agg):
        """Add a term factory"""
        self.aggs.append(agg)

    def reset(self):
        """Reset the aggs"""
        self.aggs = []

    def serialize(self):
        res = {}
        for agg in self.aggs:
            res.update(agg.serialize())
        return res


class Agg(EqualityComparableUsingAttributeDictionary):

    def __init__(self, name, scope=None, nested=None, is_global=None,
                 agg_filter=None, **kwargs):
        self.name = name
        self.scope = scope
        self.nested = nested
        self.is_global = is_global
        self.agg_filter = agg_filter

    def serialize(self):
        data = {self._internal_name: self._serialize()}
        if self.scope is not None:
            data["scope"] = self.scope
        if self.nested is not None:
            data["nested"] = self.nested
        if self.is_global:
            data['global'] = self.is_global
        if self.agg_filter:
            data['agg_filter'] = self.agg_filter.serialize()
        if isinstance(self, BucketAgg):
            return {self.name: data}
        return {self.name: data}

    def _serialize(self):
        raise NotImplementedError

    @property
    def _internal_name(self):
        raise NotImplementedError

    @property
    def _name(self):
        return self.name


class BucketAgg(Agg):
    def __init__(self, name, sub_aggs=None, **kwargs):
        super(BucketAgg, self).__init__(name, **kwargs)
        self.sub_aggs = sub_aggs
        self.name = name

    def serialize(self):
        data = super(BucketAgg, self).serialize()
        sub_data = {}
        if self.sub_aggs is not None:
            for sub_agg in self.sub_aggs:
                if isinstance(sub_agg, Agg):
                    sub_data.update(sub_agg.serialize())
                else:
                    raise RuntimeError("Invalid Agg: Only Stats-Aggregations allowed as Sub-Aggregations")
            data[self.name].update({"aggs": sub_data})
        return data

    @property
    def _internal_name(self):
        raise NotImplementedError


class FilterAgg(BucketAgg):

    _internal_name = "filter"

    def __init__(self, name, filter, **kwargs):
        super(FilterAgg, self).__init__(name, **kwargs)
        self.filter = filter

    def _serialize(self):
        return self.filter.serialize()


class FiltersAgg(BucketAgg):

    _internal_name = "filters"

    def __init__(self, name, names, filters, **kwargs):
        super(FiltersAgg, self).__init__(name, **kwargs)
        self.filters = filters
        self.names = names

    def _serialize(self):
        result = {'filters': {}}
        for name, filter in zip(self.names, self.filters):
            result['filters'][name] = filter.serialize()

        return result


class HistogramAgg(BucketAgg):

    _internal_name = "histogram"

    def __init__(self, name, field=None, interval=None, time_interval=None,
                 key_field=None, value_field=None, key_script=None, min_doc_count=None,
                 value_script=None, params=None, extended_bounds=None, **kwargs):
        super(HistogramAgg, self).__init__(name, **kwargs)
        self.field = field
        self.interval = interval
        self.min_doc_count = int(min_doc_count) if min_doc_count else None
        self.time_interval = time_interval
        self.key_field = key_field
        self.value_field = value_field
        self.key_script = key_script
        self.value_script = value_script
        self.params = params
        self.extended_bounds = extended_bounds

    def _add_interval(self, data):
        if self.interval:
            data['interval'] = self.interval
        elif self.time_interval:
            data['time_interval'] = self.time_interval
        else:
            raise RuntimeError("Invalid field: interval or time_interval required")

    def _serialize(self):
        data = {}
        if self.field:
            data['field'] = self.field
            self._add_interval(data)
        elif self.key_field:
            data['key_field'] = self.key_field
            if self.value_field:
                data['value_field'] = self.value_field
            else:
                raise RuntimeError("Invalid key_field: value_field required")
            self._add_interval(data)
        elif self.key_script:
            data['key_script'] = self.key_script
            if self.value_script:
                data['value_script'] = self.value_script
            else:
                raise RuntimeError("Invalid key_script: value_script required")
            if self.params:
                data['params'] = self.params
            if self.interval:
                data['interval'] = self.interval
            elif self.time_interval:
                data['time_interval'] = self.time_interval
        if self.min_doc_count is not None:
            data['min_doc_count'] = self.min_doc_count
        if self.extended_bounds is not None:
            data['extended_bounds'] = self.extended_bounds
        return data


class DateHistogramAgg(BucketAgg):

    _internal_name = "date_histogram"

    def __init__(self, name, field=None, interval=None, time_zone=None, pre_zone=None,
                 post_zone=None, factor=None, pre_offset=None, post_offset=None,
                 key_field=None, value_field=None, value_script=None, params=None,
                 min_doc_count=None, extended_bounds=None, **kwargs):
        super(DateHistogramAgg, self).__init__(name, **kwargs)
        self.field = field
        self.interval = interval
        self.time_zone = time_zone
        self.pre_zone = pre_zone
        self.post_zone = post_zone
        self.factor = factor
        self.pre_offset = pre_offset
        self.post_offset = post_offset
        self.key_field = key_field
        self.value_field = value_field
        self.value_script = value_script
        self.min_doc_count = int(min_doc_count) if min_doc_count else None
        self.params = params
        self.extended_bounds = extended_bounds

    def _serialize(self):
        data = {}
        if self.interval:
            data['interval'] = self.interval
        else:
            raise RuntimeError("interval required")
        if self.time_zone:
            data['time_zone'] = self.time_zone
        if self.pre_zone:
            data['pre_zone'] = self.pre_zone
        if self.post_zone:
            data['post_zone'] = self.post_zone
        if self.factor:
            data['factor'] = self.factor
        if self.pre_offset:
            data['pre_offset'] = self.pre_offset
        if self.post_offset:
            data['post_offset'] = self.post_offset
        if self.min_doc_count is not None:
            data['min_doc_count'] = self.min_doc_count
        if self.extended_bounds is not None:
            data['extended_bounds'] = self.extended_bounds
        if self.field:
            data['field'] = self.field
        elif self.key_field:
            data['key_field'] = self.key_field
            if self.value_field:
                data['value_field'] = self.value_field
            elif self.value_script:
                data['value_script'] = self.value_script
                if self.params:
                    data['params'] = self.params
            else:
                raise RuntimeError("Invalid key_field: value_field or value_script required")
        return data


class NestedAgg(BucketAgg):
    _internal_name = "nested"

    def __init__(self, name, path, **kwargs):
        super(NestedAgg, self).__init__(name, **kwargs)
        self.path = path

    def _serialize(self):
        data = {}
        data['path'] = self.path
        return data


class RangeAgg(BucketAgg):

    _internal_name = "range"

    def __init__(self, name, field=None, ranges=None, key_field=None, value_field=None,
                 key_script=None, value_script=None, params=None, **kwargs):
        super(RangeAgg, self).__init__(name, **kwargs)
        self.field = field
        self.ranges = ranges or []
        self.key_field = key_field
        self.value_field = value_field
        self.key_script = key_script
        self.value_script = value_script
        self.params = params

    def _serialize(self):
        data = {}
        if not self.ranges:
            raise RuntimeError("Invalid ranges")
        data['ranges'] = self.ranges
        if self.field:
            data['field'] = self.field
        elif self.key_field:
            data['key_field'] = self.key_field
            if self.value_field:
                data['value_field'] = self.value_field
            else:
                raise RuntimeError("Invalid key_field: value_field required")
        elif self.key_script:
            data['key_script'] = self.key_script
            if self.value_script:
                data['value_script'] = self.value_script
            else:
                raise RuntimeError("Invalid key_script: value_script required")
            if self.params:
                data['params'] = self.params
        return data


class StatsAgg(Agg):

    _internal_name = "stats"

    def __init__(self, name, field=None, script=None, params=None, **kwargs):
        super(StatsAgg, self).__init__(name, **kwargs)
        self.field = field
        self.script = script
        self.params = params

    def _serialize(self):
        data = {}
        if self.field:
            data['field'] = self.field
        elif self.script:
            data['script'] = self.script
            if self.params:
                data['params'] = self.params
        return data


class ValueCountAgg(Agg):

    _internal_name = "value_count"

    def __init__(self, name, field=None, script=None, params=None, **kwargs):
        super(ValueCountAgg, self).__init__(name, **kwargs)
        self.field = field
        self.script = script
        self.params = params

    def _serialize(self):
        data = {}
        if self.field:
            data['field'] = self.field
        elif self.script:
            data['script'] = self.script
            if self.params:
                data['params'] = self.params
        return data


class SumAgg(Agg):

    _internal_name = "sum"

    def __init__(self, name, field=None, script=None, params=None, **kwargs):
        super(SumAgg, self).__init__(name, **kwargs)
        self.field = field
        self.script = script
        self.params = params

    def _serialize(self):
        data = {}
        if self.field:
            data['field'] = self.field
        elif self.script:
            data['script'] = self.script
            if self.params:
                data['params'] = self.params
        return data


class AvgAgg(Agg):

    _internal_name = "avg"

    def __init__(self, name, field=None, script=None, params=None, **kwargs):
        super(AvgAgg, self).__init__(name, **kwargs)
        self.field = field
        self.script = script
        self.params = params

    def _serialize(self):
        data = {}
        if self.field:
            data['field'] = self.field
        elif self.script:
            data['script'] = self.script
            if self.params:
                data['params'] = self.params
        return data


class TermsAgg(BucketAgg):

    _internal_name = "terms"

    def __init__(self, name, field=None, fields=None, size=100, order=None,
                 exclude=None, regex=None, regex_flags="DOTALL", script=None,
                 lang=None, all_terms=None, min_doc_count=None, **kwargs):
        super(TermsAgg, self).__init__(name, **kwargs)
        self.field = field
        self.fields = fields
        self.size = size
        self.order = order
        self.exclude = exclude or []
        self.regex = regex
        self.regex_flags = regex_flags
        self.script = script
        self.lang = lang
        self.all_terms = all_terms
        self.min_doc_count = int(min_doc_count) if min_doc_count else None

class CardinalityAgg(Agg):

    _internal_name = "cardinality"

    def __init__(self, name, field=None, precision_threshold=100, **kwargs):
        super(CardinalityAgg, self).__init__(name, **kwargs)
        self.field = field
        if precision_threshold > 40000:
            precision_threshold = 40000
        self.precision_threshold = precision_threshold


    def _serialize(self):
        if not self.field:
            raise RuntimeError("Field is required:%s" % self.order)

        data = {}
        data['field'] = self.field
        data['precision_threshold'] = self.precision_threshold
        return data


class TermStatsAgg(Agg):

    _internal_name = "terms_stats"

    def __init__(self, name, size=10, order=None, key_field=None, value_field=None,
                 key_script=None, value_script=None, params=None, **kwargs):
        super(TermStatsAgg, self).__init__(name, **kwargs)
        self.size = size
        self.ORDER_VALUES = ['term', 'reverse_term', 'count', 'reverse_count',
                             'total', 'reverse_total', 'min', 'reverse_min',
                             'max', 'reverse_max', 'mean', 'reverse_mean']
        self.order = order if order is not None else self.ORDER_VALUES[0]
        self.key_field = key_field
        self.value_field = value_field
        self.key_script = key_script
        self.value_script = value_script
        self.params = params

    def _serialize(self):
        data = {}
        if self.size is not None:
            data['size'] = self.size
        if self.order:
            if self.order not in self.ORDER_VALUES:
                raise RuntimeError("Invalid order value:%s" % self.order)
            data['order'] = self.order

        if self.key_field:
            data['key_field'] = self.key_field
        else:
            raise RuntimeError("key_field required")

        if self.value_field:
            data['value_field'] = self.value_field
        elif self.value_script:
            data['value_script'] = self.value_script
            if self.params:
                data['params'] = self.params
        else:
            raise RuntimeError("Invalid value: value_field OR value_script required")

        return data


class AggQueryWrap(EqualityComparableUsingAttributeDictionary):

    def __init__(self, wrap_object, **kwargs):
        """Base Object for every Filter Object"""
        self.wrap_object = wrap_object

    def serialize(self):
        return {"query": self.wrap_object.serialize()}


class MissingAgg(Agg):

    _internal_name = "missing"

    def __init__(self, name, field=None, **kwargs):
        super(MissingAgg, self).__init__(name, **kwargs)
        self.field = field

    def _serialize(self):
        data = {}
        if self.field:
            data['field'] = self.field
        return data


class MinAgg(ValueCountAgg):

    _internal_name = "min"


class MaxAgg(ValueCountAgg):

    _internal_name = "max"


class ReverseNestedAgg(BucketAgg):
    _internal_name = "reverse_nested"

    def __init__(self, name, path=None, **kwargs):
        self.path = path
        super(ReverseNestedAgg, self).__init__(name=name, **kwargs)

    def _serialize(self):
        return {"path": self.path} if self.path else {}



#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import print_function

import sys
import pprint
from urlparse import urlparse
from thrift.transport import TTransport
from thrift.transport import TSocket
from thrift.transport import THttpClient
from thrift.protocol import TBinaryProtocol

from .Rest import Client
from .ttypes import *

pp = pprint.PrettyPrinter(indent = 4)
host = '127.0.0.1'
port = 9500
uri = ''
framed = False
http = False
argi = 1

socket = TSocket.TSocket(host, port)
transport = TTransport.TBufferedTransport(socket)
protocol = TBinaryProtocol.TBinaryProtocol(transport)
client = Client(protocol)
transport.open()

res = RestRequest(0, "/test-index/test-type/1", {}, {})
print(client.execute(res))

transport.close()



# -*- coding: utf-8 -*-

class SettingsBuilder(object):
    def __init__(self, settings=None, mappings=None):
        self.settings = settings or {'index.number_of_replicas': 1,
                                     "index.number_of_shards": 5}
        self.mappings = {}
        if mappings:
            self.add_mapping(mappings)

    def add_mapping(self, data, name=None):
        """
        Add a new mapping
        """
        from .mappings import DocumentObjectField
        from .mappings import NestedObject
        from .mappings import ObjectField

        if isinstance(data, (DocumentObjectField, ObjectField, NestedObject)):
            self.mappings[data.name] = data.as_dict()
            return

        if name:
            self.mappings[name] = data
            return
            if isinstance(data, dict):
                self.mappings.update(data)
            elif isinstance(data, list):
                for d in data:
                    if isinstance(d, dict):
                        self.mappings.update(d)
                    elif isinstance(d, DocumentObjectField):
                        self.mappings[d.name] = d.as_dict()
                    else:
                        name, data = d
                        self.add_mapping(data, name)

    def as_dict(self):
        """Returns a dict"""
        return {"settings": self.settings,
                "mappings": self.mappings}


class StatusProcessor(object):
    def __init__(self, connection):
        self.connection = connection
        self.cstate = connection.cluster_state()
        self.cstatus = connection.status()
        self.cnodes = connection.cluster_nodes()
        nodes = [({"code":k, "name":v["name"]}, v["transport_address"]) for k, v in self.cstate.nodes.items()]
        nodes = sorted(nodes, key=lambda v: v[0]["name"])
        self.node_names = [k for k, v in nodes]

    def get_indices_data(self):
        indices_names = sorted(self.cstatus.indices.keys())
        data = {}
        for indexname in indices_names:
            data['name'] = indexname
            data['data'] = self.cstatus.indices[indexname]
            nodes = []
            for nodename in self.node_names:
                shards = []
                for shardid, shards_replica in sorted(data['data']['shards'].items()):
                    for shard in shards_replica:
                        if shard["routing"]['node'] == nodename["code"]:
                            shards.append((shardid, shard))
                nodes.append((nodename["name"], shards))
            data['nodes'] = nodes
            yield data
            data = {}



# -*- coding: utf-8 -*-
from .exceptions import InvalidSortOrder
from .utils import EqualityComparableUsingAttributeDictionary


class SortOrder(EqualityComparableUsingAttributeDictionary):
    """
    Defines sort order
    """
    MODE_MIN = 'min'
    MODE_MAX = 'max'
    MODE_SUM = 'sum'  # not available for geo sorting
    MODE_AVG = 'avg'
    MODES = (MODE_MIN, MODE_MAX, MODE_SUM, MODE_AVG)

    def __init__(self, field=None, order=None, mode=None, nested_path=None,
                 nested_filter=None, missing=None, ignore_unmapped=None):
        self.field = field
        self.mode = mode
        self.order = order
        self.nested_path = nested_path
        self.nested_filter = nested_filter
        self.missing = missing
        self.ignore_unmapped = ignore_unmapped

    def serialize_order_params(self):
        res = {}
        if self.order:
            res['order'] = self.order
        if self.mode:
            res['mode'] = self.mode
        if self.nested_path:
            res['nested_path'] = self.nested_path
        if self.nested_filter:
            res['nested_filter'] = self.nested_filter.serialize()
        if self.missing:
            res['missing'] = self.missing
        if self.ignore_unmapped is not None:
            res['ignore_unmapped'] = self.ignore_unmapped

        return res

    def serialize(self):
        """Serialize the search to a structure as passed for a search body."""
        if not self.field:
            raise InvalidSortOrder('sort order must contain field name')

        return {self.field: self.serialize_order_params()}

    def __repr__(self):
        return str(self.serialize())


class GeoSortOrder(SortOrder):

    def __init__(self, lat=None, lon=None, geohash=None, unit=None,
                 **kwargs):
            super(GeoSortOrder, self).__init__(**kwargs)
            self.lat = lat
            self.lon = lon
            self.geohash = geohash
            self.unit = unit

    def serialize_order_params(self):
        res = super(GeoSortOrder, self).serialize_order_params()
        if self.geohash:
            res[self.field] = self.geohash
        elif self.lat is not None and self.lon is not None:
            res[self.field] = [self.lat, self.lon]
        else:
            raise InvalidSortOrder('Either geohash or lat and lon must be set')
        if self.unit:
            res['unit'] = self.unit

        return res

    def serialize(self):
        res = {
            '_geo_distance': self.serialize_order_params()
        }

        return res


class ScriptSortOrder(SortOrder):

    def __init__(self, script, type=None, params=None, **kwargs):
        super(ScriptSortOrder, self).__init__(**kwargs)
        self.script = script
        self.type = type
        self.params = params

    def serialize(self):
        res = {
            'script': self.script
        }
        if self.type:
            res['type'] = self.type
        if self.params:
            res['params'] = self.params
        if self.order:
            res['order'] = self.order

        res = {'_script': res}
        return res


class SortFactory(EqualityComparableUsingAttributeDictionary):
    """
    Container for SortOrder objects
    """

    def __init__(self):
        self.sort_orders = []

    def __bool__(self):
            return bool(self.sort_orders)

    def serialize(self):
        """Serialize the search to a structure as passed for a search body."""
        res = []
        for _sort in self.sort_orders:
            res.append(_sort.serialize())
        return res or None

    def __repr__(self):
        return str(self.serialize())

    def add(self, sort_order):
        """Add sort order"""
        self.sort_orders.append(sort_order)

    def reset(self):
        """Reset sort orders"""
        self.sort_orders = []



class River(object):
    def __init__(self, index_name=None, index_type=None, bulk_size=100, bulk_timeout=None):
        self.name = index_name
        self.index_name = index_name
        self.index_type = index_type
        self.bulk_size = bulk_size
        self.bulk_timeout = bulk_timeout

    def serialize(self):
        res = self._serialize()
        index = {}
        if self.name:
            index['name'] = self.name
        if self.index_name:
            index['index'] = self.index_name
        if self.index_type:
            index['type'] = self.index_type
        if self.bulk_size:
            index['bulk_size'] = self.bulk_size
        if self.bulk_timeout:
            index['bulk_timeout'] = self.bulk_timeout
        if index:
            res['index'] = index
        return res

    def __repr__(self):
        return str(self.serialize())

    def _serialize(self):
        raise NotImplementedError


class RabbitMQRiver(River):
    type = "rabbitmq"

    def __init__(self, host="localhost", port=5672, user="guest",
                 password="guest", vhost="/", queue="es", exchange="es",
                 routing_key="es", exchange_declare=True, exchange_type="direct",
                 exchange_durable=True, queue_declare=True, queue_durable=True,
                 queue_auto_delete=False, queue_bind=True, **kwargs):
        super(RabbitMQRiver, self).__init__(**kwargs)
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.vhost = vhost
        self.queue = queue
        self.exchange = exchange
        self.routing_key = routing_key
        self.exchange_declare = exchange_declare
        self.exchange_type = exchange_type
        self.exchange_durable = exchange_durable
        self.queue_declare = queue_declare
        self.queue_durable = queue_durable
        self.queue_auto_delete = queue_auto_delete
        self.queue_bind = queue_bind

    def _serialize(self):
        return {
            "type": self.type,
            self.type: {
                "host": self.host,
                "port": self.port,
                "user": self.user,
                "pass": self.password,
                "vhost": self.vhost,
                "queue": self.queue,
                "exchange": self.exchange,
                "routing_key": self.routing_key,
                "exchange_declare": self.exchange_declare,
                "exchange_type": self.exchange_type,
                "exchange_durable": self.exchange_durable,
                "queue_declare": self.queue_declare,
                "queue_durable": self.queue_durable,
                "queue_auto_delete": self.queue_auto_delete,
                "queue_bind": self.queue_bind
            }
        }


class TwitterRiver(River):
    type = "twitter"

    def __init__(self, user=None, password=None, **kwargs):
        self.user = user
        self.password = password
        self.consumer_key = kwargs.pop('consumer_key', None)
        self.consumer_secret = kwargs.pop('consumer_secret', None)
        self.access_token = kwargs.pop('access_token', None)
        self.access_token_secret = kwargs.pop('access_token_secret', None)
        # These filters may be lists or comma-separated strings of values
        self.tracks = kwargs.pop('tracks', None)
        self.follow = kwargs.pop('follow', None)
        self.locations = kwargs.pop('locations', None)
        super(TwitterRiver, self).__init__(**kwargs)

    def _serialize(self):
        result = {"type": self.type}
        if self.user and self.password:
            result[self.type] = {"user": self.user,
                                 "password": self.password}
        elif (self.consumer_key and self.consumer_secret and self.access_token
              and self.access_token_secret):
            result[self.type] = {"oauth": {
                "consumer_key": self.consumer_key,
                "consumer_secret": self.consumer_secret,
                "access_token": self.access_token,
                "access_token_secret": self.access_token_secret,
            }
            }
        else:
            raise ValueError("Twitter river requires authentication by username/password or OAuth")
        filter = {}
        if self.tracks:
            filter['tracks'] = self.tracks
        if self.follow:
            filter['follow'] = self.follow
        if self.locations:
            filter['locations'] = self.locations
        if filter:
            result[self.type]['filter'] = filter
        return result


class CouchDBRiver(River):
    type = "couchdb"

    def __init__(self, host="localhost", port=5984, db="mydb", filter=None,
                 filter_params=None, script=None, user=None, password=None,
                 **kwargs):
        super(CouchDBRiver, self).__init__(**kwargs)
        self.host = host
        self.port = port
        self.db = db
        self.filter = filter
        self.filter_params = filter_params
        self.script = script
        self.user = user
        self.password = password

    def serialize(self):
        result = {
            "type": self.type,
            self.type: {
                "host": self.host,
                "port": self.port,
                "db": self.db,
                "filter": self.filter,
            }
        }
        if self.filter_params is not None:
            result[self.type]["filter_params"] = self.filter_params
        if self.script is not None:
            result[self.type]["script"] = self.script
        if self.user is not None:
            result[self.type]["user"] = self.user
        if self.password is not None:
            result[self.type]["password"] = self.password
        return result


class JDBCRiver(River):
    type = "jdbc"

    def __init__(self, dbhost="localhost", dbport=5432, dbtype="postgresql",
                 dbname=None, dbuser=None, dbpassword=None, poll_time="5s",
                 sql="", name=None, params=None, **kwargs):
        super(JDBCRiver, self).__init__(**kwargs)
        self.dbsettings = {
            'dbhost': dbhost,
            'dbport': dbport,
            'dbtype': dbtype,
            'dbname': dbname,
            'dbuser': dbuser,
            'dbpassword': dbpassword,
        }
        self.poll_time = poll_time
        self.sql = sql
        self.params = params or {}
        if name is not None:
            self.name = name

    def _serialize(self):
        ret = {
            "type": self.type,
            self.type: {
                "driver": "org.%(dbtype)s.Driver" % self.dbsettings,
                "url": "jdbc:%(dbtype)s://%(dbhost)s:%(dbport)s/%(dbname)s" \
                       % self.dbsettings,
                "user": "%(dbuser)s" % self.dbsettings,
                "password": "%(dbpassword)s" % self.dbsettings,
                "strategy": "simple",
                "poll": self.poll_time,
                "sql": self.sql.replace('\n', ' '),
            }
        }

        ret.update(self.params)

        return ret


class MongoDBRiver(River):
    type = "mongodb"

    def __init__(self, servers, db, collection, index_name, mapping_type, gridfs=False, options=None, bulk_size=1000,
                 filter=None, script=None, **kwargs):
        super(MongoDBRiver, self).__init__(**kwargs)
        self.name = index_name
        self.index_type = mapping_type
        self.bulk_size = bulk_size
        self.mongodb = {
            "servers": servers,
            "db": db,
            "collection": collection,
            "options": options or {},
            "gridfs": gridfs,
            "filter": filter
        }
        if script:
          self.mongodb['script'] = script

    def _serialize(self):
        result = {
            'type': self.type,
            'mongodb': self.mongodb
        }
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    pyes.utils.compat
    ~~~~~~~~~~~~~~~~~

    Taken from celery.utils.compat
    Backward compatible implementations of features
    only available in newer Python versions.

    :copyright: (c) 2009 - 2012 by Ask Solem.
    :license: BSD, see LICENSE for more details.

"""
from __future__ import absolute_import

############## py3k #########################################################
import sys, types
is_py3k = sys.version_info[0] == 3

try:
    reload = reload                         # noqa
except NameError:                           # pragma: no cover
    from imp import reload                  # noqa

try:
    from UserList import UserList           # noqa
except ImportError:                         # pragma: no cover
    from collections import UserList        # noqa

try:
    from UserDict import UserDict           # noqa
except ImportError:                         # pragma: no cover
    from collections import UserDict        # noqa

if is_py3k:                                 # pragma: no cover
    from io import StringIO, BytesIO
    from .encoding import bytes_to_str

    class WhateverIO(StringIO):

        def write(self, data):
            StringIO.write(self, bytes_to_str(data))
else:
    from StringIO import StringIO           # noqa
    BytesIO = WhateverIO = StringIO         # noqa

if is_py3k:
    string_types = str,
    integer_types = int,
    class_types = type,
    text_type = str
    binary_type = bytes
else:
    string_types = basestring,
    integer_types = int, long
    class_types = type, types.ClassType
    text_type = unicode
    binary_type = str

############## itertools.zip_longest #######################################

try:
    from itertools import izip_longest as zip_longest
except ImportError:                         # pragma: no cover
    import itertools

    def zip_longest(*args, **kwds):  # noqa
        fillvalue = kwds.get("fillvalue")

        def sentinel(counter=([fillvalue] * (len(args) - 1)).pop):
            yield counter()     # yields the fillvalue, or raises IndexError

        fillers = itertools.repeat(fillvalue)
        iters = [itertools.chain(it, sentinel(), fillers)
                    for it in args]
        try:
            for tup in itertools.izip(*iters):
                yield tup
        except IndexError:
            pass


############## itertools.chain.from_iterable ################################
from itertools import chain


def _compat_chain_from_iterable(iterables):  # pragma: no cover
    for it in iterables:
        for element in it:
            yield element

try:
    chain_from_iterable = getattr(chain, "from_iterable")
except AttributeError:   # pragma: no cover
    chain_from_iterable = _compat_chain_from_iterable


############## logging.handlers.WatchedFileHandler ##########################
import logging
import os
from stat import ST_DEV, ST_INO
import platform as _platform

if _platform.system() == "Windows":  # pragma: no cover
    #since windows doesn't go with WatchedFileHandler use FileHandler instead
    WatchedFileHandler = logging.FileHandler
else:
    try:
        from logging.handlers import WatchedFileHandler
    except ImportError:  # pragma: no cover
        class WatchedFileHandler(logging.FileHandler):  # noqa
            """
            A handler for logging to a file, which watches the file
            to see if it has changed while in use. This can happen because of
            usage of programs such as newsyslog and logrotate which perform
            log file rotation. This handler, intended for use under Unix,
            watches the file to see if it has changed since the last emit.
            (A file has changed if its device or inode have changed.)
            If it has changed, the old file stream is closed, and the file
            opened to get a new stream.

            This handler is not appropriate for use under Windows, because
            under Windows open files cannot be moved or renamed - logging
            opens the files with exclusive locks - and so there is no need
            for such a handler. Furthermore, ST_INO is not supported under
            Windows; stat always returns zero for this value.

            This handler is based on a suggestion and patch by Chad J.
            Schroeder.
            """
            def __init__(self, *args, **kwargs):
                logging.FileHandler.__init__(self, *args, **kwargs)

                if not os.path.exists(self.baseFilename):
                    self.dev, self.ino = -1, -1
                else:
                    stat = os.stat(self.baseFilename)
                    self.dev, self.ino = stat[ST_DEV], stat[ST_INO]

            def emit(self, record):
                """
                Emit a record.

                First check if the underlying file has changed, and if it
                has, close the old stream and reopen the file to get the
                current stream.
                """
                if not os.path.exists(self.baseFilename):
                    stat = None
                    changed = 1
                else:
                    stat = os.stat(self.baseFilename)
                    changed = ((stat[ST_DEV] != self.dev) or
                               (stat[ST_INO] != self.ino))
                if changed and self.stream is not None:
                    self.stream.flush()
                    self.stream.close()
                    self.stream = self._open()
                    if stat is None:
                        stat = os.stat(self.baseFilename)
                    self.dev, self.ino = stat[ST_DEV], stat[ST_INO]
                logging.FileHandler.emit(self, record)


############## format(int, ',d') ##########################

if sys.version_info >= (2, 7):
    def format_d(i):
        return format(i, ',d')
else:
    def format_d(i):  # noqa
        s = '%d' % i
        groups = []
        while s and s[-1].isdigit():
            groups.append(s[-3:])
            s = s[:-3]
        return s + ','.join(reversed(groups))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The main QuerySet implementation. This provides the public API for the ORM.
"""

import copy
from pyes.es import ResultSetList, EmptyResultSet, ResultSet
from pyes.filters import ANDFilter, ORFilter, NotFilter, Filter, TermsFilter, TermFilter, RangeFilter, ExistsFilter, RegexTermFilter, IdsFilter, PrefixFilter, MissingFilter
from pyes.facets import Facet, TermFacet, StatisticalFacet, HistogramFacet, DateHistogramFacet, TermStatsFacet
from pyes.query import MatchAllQuery, BoolQuery, FilteredQuery, Search
from pyes.utils import ESRange, make_id
from pyes.exceptions import NotFoundException
try:
    from django.db.models import Q
except ImportError:
    class Q:pass

import six
from pyes_engine import logger

from brainaetic.documental import get_model_from_str
import re
from django.utils.functional import SimpleLazyObject

REPR_OUTPUT_SIZE = 20

# Calculate the verbose_name by converting from InitialCaps to "lowercase with spaces".
get_verbose_name = lambda class_name: re.sub('(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', ' \\1',
                                             class_name).lower().strip()
FIELD_SEPARATOR = "__"

from pyes.scriptfields import ScriptField
from pyes.exceptions import InvalidParameter


class QuerySet(object):
    """
    Represents a lazy database lookup for a set of objects.
    """

    def __init__(self, model=None, using=None, connection=None):
        #TODO: dfix
        self.model = model
        # EmptyQuerySet instantiates QuerySet with model as None
        self._using = using
        self._index = using
        self._size = 10
        self._type = None
        self._connection = connection
        self._queries = []
        self._filters = []
        self._facets = []
        self._ordering = []
        self._scan = None
        self._rescorer = None
        self._fields = [] #fields to load
        self._result_cache = None #hold the resultset
        if model:
            self._using = model._meta.using
            if connection:
                self._index = connection.database
            if model._meta.ordering:
                self._insert_ordering(self, *model._meta.ordering)

    def get_queryset(self):
        return self

    def using(self, using):
        self._using = using
        return self

    def scan(self, active=True):
        self._scan = active
        return self

    def _clear_ordering(self):
        #reset ordering
        self._ordering = []

    @property
    def type(self):
        if self._type:
            return self._type
        return self.model._meta.db_table

    def _next_is_sticky(self):
        return self

    def select_related(self, *fields, **kwargs):
        return self

        ########################

    # PYTHON MAGIC METHODS #
    ########################

    def __deepcopy__(self, memo):
        """
        Deep copy of a QuerySet doesn't populate the cache
        """
        obj = self.__class__()
        for k, v in self.__dict__.items():
            if k in ('_iter', '_result_cache'):
                obj.__dict__[k] = None
            else:
                obj.__dict__[k] = copy.deepcopy(v, memo)
        return obj

    def __getstate__(self):
        """
        Allows the QuerySet to be pickled.
        """
        # Force the cache to be fully populated.
        len(self)

        obj_dict = self.__dict__.copy()
        obj_dict['_iter'] = None
        return obj_dict

    def __repr__(self):
        data = list(self[:REPR_OUTPUT_SIZE + 1])
        if len(data) > REPR_OUTPUT_SIZE:
            data[-1] = "...(remaining elements truncated)..."
        return repr(data)

    def _build_query(self):
        if not self._queries and not self._filters:
            return MatchAllQuery()
        query = MatchAllQuery()
        if self._queries:
            if len(self._queries) == 1:
                query = self._queries[0]
            else:
                query = BoolQuery(must=self._queries)
        if not self._filters:
            return query
        if len(self._filters) == 1:
            return FilteredQuery(query, self._filters[0])
        return FilteredQuery(query, ANDFilter(self._filters))

    def _build_search(self):
        query = Search(self._build_query(), version=True)

        if self._ordering:
            query.sort = self._ordering
        if self._facets:
            for facet in self._facets:
                query.facet.add(facet)
        if self._size==0:
            query.size=0
        if self._rescorer:
            query.rescore=self._rescorer

        return query

    def _build_model(self, m, record):
        from django.db import models

        data = record["_source"]
        data["id"] = record["_id"]
        fields = self.model._meta.fields
        for field in fields:
            if field.attname not in data:
                continue

            if isinstance(field, models.ListField):
                model = field.item_field.model
                if isinstance(model, six.string_types):
                    model = get_model_from_str(model)
                if isinstance(field.item_field, models.ForeignKey):
                    #data[field.attname] = ModelIteratorResolver(field.item_field.model, data[field.attname])
                    if data[field.attname]:
                        data[field.attname] = [model.objects.get(pk=id) for id in data[field.attname]]
                    else:
                        data[field.attname] = []
            elif isinstance(field, models.SetField):
                model = field.item_field.model
                if isinstance(model, six.string_types):
                    model = get_model_from_str(model)
                if isinstance(field.item_field, models.ForeignKey):
                    if data[field.attname]:
                    #data[field.attname] = ModelIteratorResolver(field.item_field.model, data[field.attname])
                        data[field.attname] = set([model.objects.get(pk=id) for id in data[field.attname]])
                    else:
                        data[field.attname] = []
                else:
                    data[field.attname] = set(data[field.attname])
        value = self.model(**data)
        setattr(value, "_index", record["_index"])
        setattr(value, "_type", record["_type"])

        if "_score" in record:
            setattr(value, "_score", record["_score"])
        if "sort" in record:
            setattr(value, "_sort", record["sort"])
        return value

    def _is_valid_scan(self):
        """
        Detect if a type
        :return: a boolean
        """
        if self._scan is not None:
            return self._scan
        if self._ordering:
            return False
        if self._facets:
            return False
            #TODO check for child or nested query
        return True

    def _do_query(self):
        connection = self.model._meta.dj_connection
        index = self.get_index(connection)
        from pyes.es import ES

        if not isinstance(connection.connection, ES):
            print("Strange connection object", connection)
            #refresh data
        if connection._dirty:
            connection.connection.refresh(indices=index)
            connection._dirty = False

        #check ids
        if not self._facets and not self._queries and len(self._filters) == 1 and isinstance(self._filters[0],
                                                                                             IdsFilter):
            filter = self._filters[0]
            if len(filter.values) == 1:
                try:
                    return ResultSetList(items=[
                        connection.connection.get(index, self.type, str(filter.values[0]),
                                                  model=self._build_model)],
                                         model=self._build_model)
                except NotFoundException:
                    return EmptyResultSet()

        search = self._build_search()
        # if isinstance(search.query, FilteredQuery) and isinstance(search.query.query, MatchAllQuery) and \
        #     isinstance(search.query.filter, IdsFilter) and len(search.query.filter.values) == 1:
        #     results = [connection.connection.get(index, self.type,search.query.filter.values[0], model=self._build_model)]
        #     #iterator

        # show query
        scan = self._is_valid_scan()
        if scan:
            search.sort = {}
        #print "scan?", scan, index, self.type, search.serialize()
        return connection.connection.search(search, indices=index, doc_types=self.type,
                                            model=self._build_model, scan=scan)


    def __len__(self):
        # Since __len__ is called quite frequently (for example, as part of
        # list(qs), we make some effort here to be as efficient as possible
        # whilst not messing up any existing iterators against the QuerySet.
        if self._result_cache is None:
            self._result_cache = self._do_query()
        return self._result_cache.total

    def __iter__(self):
        if self._result_cache is None:
            len(self)
            # Python's list iterator is better than our version when we're just
        # iterating over the cache.
        return iter(self._result_cache)

    def __bool__(self):
        return self.__nonzero__()

    def __nonzero__(self):
        if self._result_cache is not None:
            len(self)
            return bool(self._result_cache.total != 0)
        try:
            next(iter(self))
        except StopIteration:
            return False
        return True

    #    def __contains__(self, val):
    #        # The 'in' operator works without this method, due to __iter__. This
    #        # implementation exists only to shortcut the creation of Model
    #        # instances, by bailing out early if we find a matching element.
    #        pos = 0
    #        if self._result_cache is not None:
    #            if val in self._result_cache:
    #                return True
    #            elif self._iter is None:
    #                # iterator is exhausted, so we have our answer
    #                return False
    #            # remember not to check these again:
    #            pos = len(self._result_cache)
    #        else:
    #            # We need to start filling the result cache out. The following
    #            # ensures that self._iter is not None and self._result_cache is not
    #            # None
    #            it = iter(self)
    #
    #        # Carry on, one result at a time.
    #        while True:
    #            if len(self._result_cache) <= pos:
    #                self._fill_cache(num=1)
    #            if self._iter is None:
    #                # we ran out of items
    #                return False
    #            if self._result_cache[pos] == val:
    #                return True
    #            pos += 1

    def __getitem__(self, k):
        """
        Retrieves an item or slice from the set of results.
        """
        if not isinstance(k, (slice,) + six.integer_types):
            raise TypeError
        assert ((not isinstance(k, slice) and (k >= 0))
                or (isinstance(k, slice) and (k.start is None or k.start >= 0)
                    and (k.stop is None or k.stop >= 0))), \
            "Negative indexing is not supported."

        if self._result_cache is None:
            len(self)
        return self._result_cache.__getitem__(k)

    def __and__(self, other):
        combined = self._clone()
        if not other._filters:
            return combined
        if other._filters:
            combined._filters.extend(other._filters)
        return combined

    def __or__(self, other):
        combined = self._clone()
        if not other._filters:
            return other._clone()
        combined._filters = ORFilter([combined._filters, other._filters])
        return combined

    ####################################
    # METHODS THAT DO DATABASE QUERIES #
    ####################################

    def iterator(self):
        """
        An iterator over the results from applying this QuerySet to the
        database.
        """
        if not self._result_cache:
            len(self)
        for r in self._result_cache:
            yield r

    def aggregate(self, *args, **kwargs):
        """
        Returns a dictionary containing the calculations (aggregation)
        over the current queryset

        If args is present the expression is passed as a kwarg using
        the Aggregate object's default alias.
        """
        obj = self._clone()
        obj._facets = [] #reset facets
        from django.db.models import Avg, Max, Min

        for arg in args:
            if isinstance(arg, (Avg, Max, Min)):

                field, djfield = self._django_to_es_field(arg.lookup)
                if not djfield:
                    obj = obj.annotate(field)
                from pyes.facets import StatisticalFacet

                obj = obj.annotate(StatisticalFacet(field, field))
        facets = obj.get_facets()
        #collecting results
        result = {}
        for name, values in facets.items():
            for k, v in values.items():
                if k.startswith("_"):
                    continue
                result[u'%s__%s' % (name, k)] = v

        return result

    def count(self):
        """
        Performs a SELECT COUNT() and returns the number of records as an
        integer.

        If the QuerySet is already fully cached this simply returns the length
        of the cached results set to avoid multiple SELECT COUNT(*) calls.
        """
        return len(self)

    def get(self, *args, **kwargs):
        """
        Performs the query and returns a single object matching the given
        keyword arguments.
        """
        clone = self.filter(*args, **kwargs)
        num = len(clone)
        if num == 1:
            return clone._result_cache[0]
        if not num:
            raise self.model.DoesNotExist(
                "%s matching query does not exist. "
                "Lookup parameters were %s" %
                #(self.model._meta.object_name, kwargs))
                (self.model.__class__.__name__, kwargs))
        raise self.model.MultipleObjectsReturned(
            "get() returned more than one %s -- it returned %s! "
            "Lookup parameters were %s" %
            #(self.model._meta.object_name, num, kwargs))
            (self.model.__class__.__name__, num, kwargs))

    def create(self, **kwargs):
        """
        Creates a new object with the given kwargs, saving it to the database
        and returning the created object.
        """
        obj = self.model(**kwargs)
        obj.save(force_insert=True, using=self._using)
        return obj

    def bulk_create(self, objs, batch_size=None):
        """
        Inserts each of the instances into the database. This does *not* call
        save() on each of the instances, does not send any pre/post save
        signals, and does not set the primary key attribute if it is an
        autoincrement field.
        """
        self._insert(objs, batch_size=batch_size, return_id=False, force_insert=True)
        self.refresh()

    def get_or_create(self, **kwargs):
        """
        Looks up an object with the given kwargs, creating one if necessary.
        Returns a tuple of (object, created), where created is a boolean
        specifying whether an object was created.
        """
        assert kwargs, \
            'get_or_create() must be passed at least one keyword argument'
        defaults = kwargs.pop('defaults', {})
        lookup = kwargs.copy()
        #TODO: check fields
        try:
            return self.get(**lookup), False
        except self.model.DoesNotExist:
            params = dict([(k, v) for k, v in kwargs.items() if '__' not in k])
            params.update(defaults)
            obj = self.model(**params)
            obj.save(force_insert=True)
            return obj, True

    def latest(self, field_name=None):
        """
        Returns the latest object, according to the model's 'get_latest_by'
        option or optional given field_name.
        """
        latest_by = field_name or "_id"#self.model._meta.get_latest_by
        assert bool(latest_by), "latest() requires either a field_name parameter or 'get_latest_by' in the model"
        obj = self._clone()
        obj.size = 1
        obj._clear_ordering()
        obj._ordering.append({latest_by: "desc"})
        return obj.get()

    def in_bulk(self, id_list):
        """
        Returns a dictionary mapping each of the given IDs to the object with
        that ID.
        """
        if not id_list:
            return {}
        qs = self._clone()
        qs.add_filter(('pk__in', id_list))
        qs._clear_ordering(force_empty=True)
        return dict([(obj._get_pk_val(), obj) for obj in qs])

    def delete(self):
        """
        Deletes the records in the current QuerySet.
        """
        del_query = self._clone()

        # The delete is actually 2 queries - one to find related objects,
        # and one to delete. Make sure that the discovery of related
        # objects is performed on the same database as the deletion.
        del_query._clear_ordering()
        connection = self.model._meta.dj_connection

        connection.connection.delete_by_query(indices=connection.database, doc_types=self.model._meta.db_table,
                                              query=del_query._build_query())
        connection.connection.indices.refresh(indices=connection.database)
        # Clear the result cache, in case this QuerySet gets reused.
        self._result_cache = None

    def update(self, *args, **kwargs):
        """
        Updates all elements in the current QuerySet, setting all the given
        fields to the appropriate values.
        """
        if args:
            for arg in args:
                if not isinstance(arg, ScriptField):
                    raise InvalidParameter("%s this must be a ScriptField or derivated" % arg)

        query = self._build_query()
        connection = self.model._meta.dj_connection
        query.fields = ["_id", "_type", "_index"]
        # results = connection.connection.search(query, indices=connection.database, doc_types=self.type,
        #                             model=self.model, scan=True)

        conn = connection.connection
        results = conn.search(query, indices=connection.database, doc_types=self.type, scan=True)

        for item in results:
            if kwargs:
                conn.update(item._meta.index, item._meta.type, item._meta.id, document=kwargs, bulk=True)
            for script in args:
                conn.update(item._meta.index, item._meta.type, item._meta.id,
                            script=script.script, lang=script.lang, params=script.params, bulk=True)

        connection.connection.flush_bulk(True)
        # Clear the result cache, in case this QuerySet gets reused.
        self._result_cache = None

    def exists(self):
        if self._result_cache is None:
            len(self)
        return bool(self._result_cache.total != 0)

    ##################################################
    # PUBLIC METHODS THAT RETURN A QUERYSET SUBCLASS #
    ##################################################

    def _django_to_es_field(self, field):
        """We use this function in value_list and ordering to get the correct fields name"""
        from django.db import models

        prefix = ""
        if field.startswith("-"):
            prefix = "-"
            field = field.lstrip("-")

        if field in ["id", "pk"]:
            return "_id", models.AutoField

        try:
            dj_field, _, _, _ = self.model._meta.get_field_by_name(field)
            if isinstance(dj_field, models.ForeignKey):
                return prefix + field + "_id", models.ForeignKey
            else:
                return prefix + field, dj_field
        except models.FieldDoesNotExist:
            pass

        return prefix + field.replace(FIELD_SEPARATOR, "."), None


    def _django_to_es_fields(self, dj_fields):
        """We use this function in value_list and ordering to get the correct fields name"""
        return [self._django_to_es_field(field)[0] for field in dj_fields]

    def _cooked_fields(self, dj_fields):
        """
        Returns a tuple of cooked fields
        :param dj_fields: a list of django name fields
        :return:
        """
        from django.db import models

        valids = []
        for field in dj_fields:
            try:
                dj_field, _, _, _ = self.model._meta.get_field_by_name(field)
                if isinstance(dj_field, models.ForeignKey):
                    valids.append((field + "_id", field, dj_field))
                else:
                    valids.append((field, field, dj_field))
            except models.FieldDoesNotExist:
                valids.append((field, field, None))
        return valids


    def values(self, *fields):
        assert fields, "A least a field is required"
        search = self._build_search()
        search.facet.reset()
        search.fields = self._django_to_es_fields(fields)
        connection = self.model._meta.dj_connection

        return connection.connection.search(search, indices=connection.database, doc_types=self.type)

    def values_list(self, *fields, **kwargs):
        flat = kwargs.pop('flat', False)
        if kwargs:
            raise TypeError('Unexpected keyword arguments to values_list: %s'
                            % (kwargs.keys(),))
        if flat and len(fields) > 1:
            raise TypeError("'flat' is not valid when values_list is called with more than one field.")
        assert fields, "A least a field is required"
        search = self._build_search()
        search.facet.reset()
        search.fields = self._django_to_es_fields(fields)
        connection = self.model._meta.dj_connection

        values_resolver = ValueListResolver(self.model, self._cooked_fields(fields), flat)
        return connection.connection.search(search, indices=connection.database, doc_types=self.type,
                                            model=values_resolver)

    def dates(self, field_name, kind, order='ASC'):
        """
        Returns a list of datetime objects representing all available dates for
        the given field_name, scoped to 'kind'.
        """

        assert kind in ("month", "year", "day", "week", "hour", "minute"), \
            "'kind' must be one of 'year', 'month', 'day', 'week', 'hour' and 'minute'."
        assert order in ('ASC', 'DESC'), \
            "'order' must be either 'ASC' or 'DESC'."

        search = self._build_search()
        search.facet.reset()
        search.facet.add_date_facet(name=field_name.replace("__", "."),
                                    field=field_name, interval=kind)
        search.size = 0
        connection = self.model._meta.dj_connection

        resulset = connection.connection.search(search, indices=connection.database, doc_types=self.type)
        resulset.fix_facets()
        entries = []
        for val in resulset.facets.get(field_name.replace("__", ".")).get("entries", []):
            if "time" in val:
                entries.append(val["time"])
        if order == "ASC":
            return sorted(entries)

        return sorted(entries, reverse=True)

    def none(self):
        """
        Returns an empty QuerySet.
        """
        return EmptyQuerySet(model=self.model, using=self._using, connection=self._connection)


    ##################################################################
    # PUBLIC METHODS THAT ALTER ATTRIBUTES AND RETURN A NEW QUERYSET #
    ##################################################################

    def all(self):
        """
        Returns a new QuerySet that is a copy of the current one. This allows a
        QuerySet to proxy for a model manager in some cases.
        """
        return self._clone()

    def query(self, *args, **kwargs):
        """
        Returns a new QuerySet instance with the args ANDed to the existing
        set.
        """
        clone = self._clone()
        queries = []
        from pyes.query import Query

        if args:
            for f in args:
                if isinstance(f, Query):
                    queries.append(f)
                else:
                    raise TypeError('Only Query objects can be passed as argument')

        for field, value in kwargs.items():
            if value == [None]:
                value = ["$$$$"]
            queries.append(self._build_inner_query(field, value))

        clone._queries.extend(queries)
        return clone

    def rescorer(self, rescorer):
        """
        Returns a new QuerySet with a set rescorer.
        """
        clone = self._clone()
        clone._rescorer=rescorer
        return clone

    def filter(self, *args, **kwargs):
        """
        Returns a new QuerySet instance with the args ANDed to the existing
        set.
        """
        return self._filter_or_exclude(False, *args, **kwargs)

    def exclude(self, *args, **kwargs):
        """
        Returns a new QuerySet instance with NOT (args) ANDed to the existing
        set.
        """
        return self._filter_or_exclude(True, *args, **kwargs)


    def _filter_or_exclude(self, negate, *args, **kwargs):
        clone = self._clone()
        filters = self._build_filter(*args, **kwargs)

        if negate:
            if len(filters) > 1:
                filters = ANDFilter(filters)
            else:
                filters = filters[0]
            clone._filters.append(NotFilter(filters))
        else:
            clone._filters.extend(filters)
        return clone

    FILTER_OPERATORS = {
        'exact': '',
        'iexact': "",
        'contains': "",
        'icontains': "",
        'regex': '',
        'iregex': "",
        'gt': '',
        'gte': '',
        'lt': '',
        'lte': '',
        'startswith': "",
        'endswith': "",
        'istartswith': "",
        'iendswith': "",
        'in': "",
        'ne': "",
        'isnull': "",
        'ismissing': "",
        'exists': "",
        'year': "",
        'prefix': "",
        'match': "",
    }

    def _get_filter_modifier(self, field):
        """Detect the filter modifier"""
        tokens = field.split(FIELD_SEPARATOR)
        if len(tokens) == 1:
            return field, ""
        if tokens[-1] in self.FILTER_OPERATORS.keys():
            return u'.'.join(tokens[:-1]), tokens[-1]
        return u'.'.join(tokens), ""

    def _prepare_value(self, value, dj_field=None):
        """Cook the value"""
        from django.db import models

        if isinstance(value, (six.string_types, int, float)):
            return value
        elif isinstance(value, SimpleLazyObject):
            return value.pk
        elif isinstance(value, models.Model):
            if dj_field:
                if isinstance(dj_field, models.ForeignKey):
                    return value.pk
                    #TODO validate other types
            return value
        elif isinstance(value, (ResultSet, EmptyQuerySet, EmptyResultSet, list)):
            return [self._prepare_value(v, dj_field=dj_field) for v in value]
        elif isinstance(value, QuerySet):
            value = value._clone()
            return [self._prepare_value(v, dj_field=dj_field) for v in value]
        return value

    def _build_inner_filter(self, field, value):
        from django.db import models


        field, modifier = self._get_filter_modifier(field)
        dj_field = None
        try:
            dj_field, _, _, _ = self.model._meta.get_field_by_name(field)
            if isinstance(dj_field, models.ForeignKey):
                field += "_id"
                if isinstance(value, models.Model):
                    if not value.pk:
                        value.save()
                    value = value.pk
            if dj_field:
                if isinstance(dj_field, models.ForeignKey) and modifier == "id":
                    field += "_id"
                    modifier = ""
                elif isinstance(dj_field, models.ListField) and isinstance(dj_field.item_field, models.ForeignKey):
                    dj_field = dj_field.item_field

        except models.FieldDoesNotExist:
            pass

        #calculating lazy resultset
        value = self._prepare_value(value, dj_field=dj_field)

        pk_column = self.model._meta.pk.column
        #case filter id
        if modifier in ["", "in"] and field in ["pk", pk_column]:
            if isinstance(value, list):
                return IdsFilter(value)
            return IdsFilter([value])

        if not modifier:
            if isinstance(value, list):
                return TermsFilter(field, value)
            return TermFilter(field, value)

        if modifier == "in":
            return TermsFilter(field, value)
        elif modifier == "gt":
            return RangeFilter(ESRange(field, from_value=value, include_lower=False))
        elif modifier == "gte":
            return RangeFilter(ESRange(field, from_value=value, include_lower=True))
        elif modifier == "lte":
            return RangeFilter(ESRange(field, to_value=value, include_upper=True))
        elif modifier == "lt":
            return RangeFilter(ESRange(field, to_value=value, include_upper=False))
        elif modifier == "in":
            return TermsFilter(field, values=value)
        elif modifier == "ne":
            return NotFilter(TermFilter(field, value))
        elif modifier == "exists":
            if isinstance(value, bool) and value:
                return ExistsFilter(field)
            return NotFilter(ExistsFilter(field))
        elif modifier in ["startswith", "istartswith"]:
            return RegexTermFilter(field, "^" + value + ".*",
                                   ignorecase=modifier == "istartswith")
        elif modifier in ["endswith", "iendswith"]:
            return RegexTermFilter(field, ".*" + value + "$",
                                   ignorecase=modifier == "iendswith")
        elif modifier in ["contains", "icontains"]:
            return RegexTermFilter(field, ".*" + value + ".*",
                                   ignorecase=modifier == "icontains")
        elif modifier in ["regex", "iregex"]:
            return RegexTermFilter(field, value,
                                   ignorecase=modifier == "iregex")
        elif modifier == "prefix":
            return PrefixFilter(field, value)

        elif modifier in ["exact", "iexact"]:
            return TermFilter(field, value)
        elif modifier == "year":
            if isinstance(value, list):
                return RangeFilter(ESRange(field, value[0], value[1], True, True))
            if isinstance(value, tuple):
                return RangeFilter(ESRange(field, value[0], value[1], True, True))
            else:
                import datetime
                from dateutil.relativedelta import relativedelta

                start = datetime.datetime(value, 1, 1)
                end = start + relativedelta(years=1)
                return RangeFilter(ESRange(field, start, end, True, False))
        elif modifier in ["isnull"]:
            if value:
                return MissingFilter(field)
            return ExistsFilter(field)

        raise NotImplementedError()

    def _build_inner_query(self, field, value):
        from django.db import models
        from pyes.query import IdsQuery, TermQuery, TermsQuery, RangeQuery, RegexTermQuery, PrefixQuery, MatchQuery

        field, modifier = self._get_filter_modifier(field)
        dj_field = None
        try:
            dj_field, _, _, _ = self.model._meta.get_field_by_name(field)
            if isinstance(dj_field, models.ForeignKey):
                field += "_id"
                if isinstance(value, models.Model):
                    if not value.pk:
                        value.save()
                    value = value.pk
            if dj_field:
                if isinstance(dj_field, models.ForeignKey) and modifier == "id":
                    field += "_id"
                    modifier = ""
                elif isinstance(dj_field, models.ListField) and isinstance(dj_field.item_field, models.ForeignKey):
                    dj_field = dj_field.item_field

        except models.FieldDoesNotExist:
            pass

        #calculating lazy resultset
        value = self._prepare_value(value, dj_field=dj_field)

        pk_column = self.model._meta.pk.column
        #case filter id
        if modifier in ["", "in"] and field in ["pk", pk_column]:
            if isinstance(value, list):
                return IdsQuery(value)
            return IdsQuery([value])

        if not modifier:
            if isinstance(value, list):
                return TermsQuery(field, value)
            return TermQuery(field, value)

        if modifier == "in":
            return TermsQuery(field, value)
        elif modifier == "gt":
            return RangeQuery(ESRange(field, from_value=value, include_lower=False))
        elif modifier == "gte":
            return RangeQuery(ESRange(field, from_value=value, include_lower=True))
        elif modifier == "lte":
            return RangeQuery(ESRange(field, to_value=value, include_upper=True))
        elif modifier == "lt":
            return RangeQuery(ESRange(field, to_value=value, include_upper=False))
        elif modifier == "in":
            return TermsQuery(field, values=value)
        elif modifier in ["startswith", "istartswith"]:
            return RegexTermQuery(field, "^" + value + ".*",
                                  ignorecase=modifier == "istartswith")
        elif modifier in ["endswith", "iendswith"]:
            return RegexTermQuery(field, ".*" + value + "$",
                                  ignorecase=modifier == "iendswith")
        elif modifier in ["contains", "icontains"]:
            return RegexTermQuery(field, ".*" + value + ".*",
                                  ignorecase=modifier == "icontains")
        elif modifier in ["regex", "iregex"]:
            return RegexTermQuery(field, value,
                                  ignorecase=modifier == "iregex")
        elif modifier == "prefix":
            return PrefixQuery(field, value)

        elif modifier in ["exact", "iexact"]:
            return TermQuery(field, value)
        elif modifier == "year":
            if isinstance(value, list):
                return RangeQuery(ESRange(field, value[0], value[1], True, True))
            if isinstance(value, tuple):
                return RangeQuery(ESRange(field, value[0], value[1], True, True))
            else:
                import datetime
                from dateutil.relativedelta import relativedelta

                start = datetime.datetime(value, 1, 1)
                end = start + relativedelta(years=1)
                return RangeQuery(ESRange(field, start, end, True, False))
        elif modifier in ["match"]:
            return MatchQuery(field, value)
        raise NotImplementedError()

    def _Q_to_filter(self, q):
        """
        Convert a Q object to filter
        :param q: a Q Object
        :return: a filter object
        """
        default_filter = ANDFilter
        if q.connector == "OR":
            default_filter = ORFilter
        filters = []
        for child in q.children:
            if isinstance(child, Q):
                if child.children:
                    filters.append(self._Q_to_filter(child))
            elif isinstance(child, tuple):
                field, value = child
                filters.append(self._build_inner_filter(field, value))
        if len(filters) == 1:
            filter = filters[0]
            if q.negated:
                return NotFilter(filter)
            return filter
        if q.negated:
            return NotFilter(default_filter(filters))
        return default_filter(filters)


    def _build_filter(self, *args, **kwargs):
        filters = []

        if args:
            for f in args:
                if isinstance(f, Filter):
                    filters.append(f)
                    #elif isinstance(f, dict):
                    #TODO: dict parser
                elif isinstance(f, Q):
                    filters.append(self._Q_to_filter(f))
                else:
                    raise TypeError('Only Filter objects can be passed as argument')

        for field, value in kwargs.items():
            if value == [None]:
                value = ["$$$$"]
                # if FIELD_SEPARATOR in field:
            #     if field.rsplit(FIELD_SEPARATOR, 1)[1] == 'isnull' and value == False:
            #         return filters
            filters.append(self._build_inner_filter(field, value))
        return filters

    def connection(self, connection):
        self._connection = connection
        return self

    def complex_filter(self, filter_obj):
        """
        Returns a new QuerySet instance with filter_obj added to the filters.

        filter_obj can be a Q object (or anything with an add_to_query()
        method) or a dictionary of keyword lookup arguments.

        This exists to support framework features such as 'limit_choices_to',
        and usually it will be more natural to use other methods.
        """
        if isinstance(filter_obj, Filter):
            clone = self._clone()
            clone._filters.add(filter_obj)
            return clone
        return self._filter_or_exclude(None, **filter_obj)


    def facet(self, *args, **kwargs):
        return self.annotate(*args, **kwargs)

    def annotate(self, *args, **kwargs):
        """
        Return a query set in which the returned objects have been annotated
        with data aggregated from related fields.
        """
        obj = self._clone()
        if args:
            for arg in args:
                if isinstance(arg, Facet):
                    obj._facets.append(arg)
                elif isinstance(arg, six.string_types):
                    modifier = "term"
                    tokens = arg.split("__")
                    if len(tokens)>1 and tokens[-1] in ["term", "stat", "histo", "date"]:
                        modifier=tokens[-1]
                        tokens=tokens[:-1]
                    field, djfield = self._django_to_es_field("__".join(tokens))
                    if modifier=="term":
                        obj._facets.append(TermFacet(name=arg, field=field, **kwargs))
                    elif modifier=="term_stat":
                        obj._facets.append(TermStatsFacet(name=arg, **kwargs))
                    elif modifier=="stat":
                        obj._facets.append(StatisticalFacet(name=arg, field=field, **kwargs))
                    elif modifier=="histo":
                        obj._facets.append(HistogramFacet(name=arg, field=field, **kwargs))
                    elif modifier=="date":
                        obj._facets.append(DateHistogramFacet(name=arg, field=field, **kwargs))
                else:
                    raise NotImplementedError("invalid type")
        else:
            # Add the aggregates/facet to the query
            for name, field in kwargs.items():
                obj._facets.append(
                    TermFacet(field=field.replace(FIELD_SEPARATOR, "."), name=name.replace(FIELD_SEPARATOR, ".")))

        return obj

    def order_by(self, *field_names):
        """
        Returns a new QuerySet instance with the ordering changed.
        We have a special field "_random"
        """
        obj = self._clone()
        obj._clear_ordering()
        self._insert_ordering(obj, *field_names)
        return obj


    def _insert_ordering(self, obj, *field_names):

        for field in field_names:
            if isinstance(field, dict):
                obj._clear_ordering()
                obj._ordering.append(field)
            elif field == "_random":
                obj._clear_ordering()
                obj._ordering.append({
                    "_script": {
                        "script": "Math.random()",
                        "type": "number",
                        "params": {},
                        "order": "asc"
                    }
                })
                break
            elif field.startswith("-"):
                obj._ordering.append({field.lstrip("-"): {"order": "desc", "ignore_unmapped": True}})
            else:
                obj._ordering.append({field: {"order": "asc", "ignore_unmapped": True}})
        return obj


    def distinct(self, *field_names):
        """
        Returns a new QuerySet instance that will select only distinct results.
        """
        return self


    def reverse(self):
        """
        Reverses the ordering of the QuerySet.
        """
        clone = self._clone()
        assert self._ordering, "You need to set an ordering for reverse"
        ordering = []
        for order in self._ordering:
            for k, v in order.items():
                if v == "asc":
                    ordering.append({k: "desc"})
                else:
                    ordering.append({k: "asc"})
        clone._ordering = ordering
        return clone

    def defer(self, *fields):
        """
        Defers the loading of data for certain fields until they are accessed.
        The set of fields to defer is added to any existing set of deferred
        fields. The only exception to this is if None is passed in as the only
        parameter, in which case all deferrals are removed (None acts as a
        reset option).
        """
        raise NotImplementedError()

    #        clone = self._clone()
    #        if fields == (None,):
    #            clone.query.clear_deferred_loading()
    #        else:
    #            clone.query.add_deferred_loading(fields)
    #        return clone

    def only(self, *fields):
        """
        Essentially, the opposite of defer. Only the fields passed into this
        method and that are not already specified as deferred are loaded
        immediately when the queryset is evaluated.
        """
        clone = self._clone()
        clone._fields = fields
        return clone

    def index(self, alias):
        """
        Selects which database this QuerySet should execute its query against.
        """
        clone = self._clone()
        clone._index = alias
        return clone

    def size(self, size):
        """
        Set the query size of  this QuerySet should execute its query against.
        """
        clone = self._clone()
        clone._size = size
        return clone

    ###################################
    # PUBLIC INTROSPECTION ATTRIBUTES #
    ###################################

    def ordered(self):
        """
        Returns True if the QuerySet is ordered -- i.e. has an order_by()
        clause or a default ordering on the model.
        """
        return len(self._ordering) > 0

    ordered = property(ordered)


    ###################
    # PUBLIC METHODS  #
    ###################

    @classmethod
    def from_qs(cls, qs, **kwargs):
        """
        Creates a new queryset using class `cls` using `qs'` data.

        :param qs: The query set to clone
        :keyword kwargs: The kwargs to pass to _clone method
        """
        assert issubclass(cls, QuerySet), "%s is not a QuerySet subclass" % cls
        assert isinstance(qs, QuerySet), "qs has to be an instance of queryset"
        return qs._clone(klass=cls, **kwargs)

    def evaluated(self):
        """
        Lets check if the queryset was already evaluated without accessing
        private methods / attributes
        """
        return not self._result_cache is None

    ###################
    # PRIVATE METHODS #
    ###################
    def _batched_insert(self, objs, fields, batch_size):
        """
        A little helper method for bulk_insert to insert the bulk one batch
        at a time. Inserts recursively a batch from the front of the bulk and
        then _batched_insert() the remaining objects again.
        """
        raise NotImplementedError()

    #        if not objs:
    #            return
    #        ops = connections[self.db].ops
    #        batch_size = (batch_size or max(ops.bulk_batch_size(fields, objs), 1))
    #        for batch in [objs[i:i+batch_size]
    #                      for i in range(0, len(objs), batch_size)]:
    #            self.model._base_manager._insert(batch, fields=fields,
    #                                             using=self.db)

    def _get_hash_id(self, record, uniq_together):
        valid_fields = []
        if isinstance(uniq_together, six.string_types):
            valid_fields.append(uniq_together)
        elif isinstance(uniq_together, (list, tuple)):
            for v in uniq_together:
                if isinstance(v, six.string_types):
                    valid_fields.append(v)
                else:
                    valid_fields.extend(list(v))
        if not valid_fields:
            return None

        values = []
        for f in valid_fields:
            val = record.get(f, u"")
            if val:
                values.append(unicode(val))
        if not values:
            return None

        return make_id("_".join(values).strip().lower())

    def _insert(self, objs, fields=None, **kwargs):

        single = False
        using = kwargs.get("using")
        if not isinstance(objs, list):
            single = True
            objs = [objs]
        if len(objs) == 1:
            single = True

        #get fields
        if not fields:
            fields = self.model._meta.concrete_fields

        connection = self.model._meta.dj_connection

        if using and using not in ["default", "echidnasearch"]:
            database = using
        else:
            database=connection.database

        meta = self.model._meta

        db_table = meta.db_table
        bulk = not kwargs.get("return_id", False)
        force_insert = kwargs.get("force_insert", False)
        pk_column = meta.pk.column
        unique_together_name = []
        valid_fields = []
        if isinstance(meta.unique_together, six.string_types):
            valid_fields.append(meta.uniq_together)
        elif isinstance(meta.unique_together, (list, tuple)):
            for v in meta.unique_together:
                if isinstance(v, six.string_types):
                    valid_fields.append(v)
                else:
                    valid_fields.extend(list(v))
                    # for field in fields:
                    #     if field.name in valid_fields:
                    #         unique_together_name.append(field.attname)

        ids = []
        for obj in objs:
            record = obj.to_dict(fields, **kwargs)
            pk = record.get(pk_column, None)
            record.pop("_id", None)
            #remove empty values
            if pk is None:
                # pk = self._get_hash_id(record, meta.unique_together)
                pk = obj.calc_pk()

            res = connection.connection.index(record, index=database, doc_type=db_table, id=pk, bulk=bulk,
                                              force_insert=force_insert)
            if not bulk:
                ids.append(res._id)
        if connection.force_refresh:
            connection.connection.refresh()
        else:
            connection._dirty = True
        if bulk:
            return []
        if single:
            return ids[0]
        return ids

    _insert.queryset_only = False

    def _clone(self, klass=None, setup=False, **kwargs):
        params = dict(model=self.model, using=self.index)
        if klass is None:
            klass = self.__class__
        if repr(klass) == "<class 'django.db.models.fields.related.RelatedManager'>":
            params["instance"] = getattr(self, "instance", None)# only for RelatedManager
        c = klass(**params)
        #copy filters/queries/facet????
        c.__dict__.update(kwargs)
        c._queries = list(self._queries)
        c._filters = list(self._filters)
        c._facets = list(self._facets)
        c._fields = list(self._fields)
        c._ordering = list(self._ordering)
        c._size = self._size
        c._scan = self._scan
        c._rescorer = self._rescorer
        c._connection = self._connection
        c._using = self._using
        return c

    # When used as part of a nested query, a queryset will never be an "always
    # empty" result.
    value_annotation = True

    def get_facets(self):
        len(self)
        if isinstance(self._result_cache, ResultSetList):
            self._result_cache = None
            len(self)
        self._result_cache.fix_facets()
        return FacetHelper(self.model, self._result_cache.facets)

    #converting functions
    #TODO: remove
    def convert_value_for_db(self, db_type, value):
        if db_type == "unicode" and not isinstance(value, six.string_types):
            return unicode(value)
        return value

    #def _build_objects(self, objs, fields=None, raw=False, **kwargs):
    #    from django.db import connections
    #
    #    connection = connections[self.model._meta.using]
    #    from django.db.models import AutoField
    #
    #    for obj in objs:
    #        data = {}
    #        for f in fields:
    #            val = f.get_db_prep_save(getattr(obj, f.attname) if raw else f.pre_save(obj, True),
    #                                     connection=connection)
    #
    #            if not f.null and val is None:
    #                if isinstance(f, AutoField):
    #                    continue
    #                raise IntegrityError("You can't set %s (a non-nullable field) to None!" % f.name)
    #
    #            db_type = f.db_type(connection=connection)
    #            value = self.convert_value_for_db(db_type, val)
    #            if value is None:
    #                continue
    #            data[f.column] = value
    #        yield (obj, data)
    #
    #    raise StopIteration

    def refresh(self):
        """Refresh an index"""
        connection = self.model._meta.dj_connection
        return connection.connection.indices.refresh(indices=connection.database)


    def terms(self, field):
        connection = self.model._meta.dj_connection
        return connection.connection.terms(connection.database, self.model._meta.db_table, field=field)["terms"]

    def _update(self, values):

        connection = self.model._meta.dj_connection
        index = self.get_index(connection)

        meta = self.model._meta
        db_table = meta.db_table
        updated = False
        for obj in self:
            record = {}
            for field, boh, new_value in values:
                record[field.attname] = field.get_db_prep_save(new_value, connection=connection)
            res = connection.connection.update(index=index, doc_type=self.type, id=obj.pk,
                                               document=record)
            updated = True
        return updated

    _update.queryset_only = False

    def get_objects_for_user(self, user, perms=None):
        if user.is_superuser:
            return self.all()
        prefix = "u.%s" % user.pk
        group_perms = user.get_group_cooked_perm_list()

        return self.filter(perms__startswith=prefix)

        # if perms is None:
        #     perms = [u'%s.%s.view' % (self.model._meta.app_label, self.model._meta.model_name)]
        #
        # return get_objects_for_user(user, perms=perms, klass=self.model)

    def get_index(self, connection=None):
        if self.model._meta.index :
            return self.model._meta.index
        if self._using and self._using not in ["default", "echidnasearch"]:
            return self._using
        return self.model._meta.get_index(connection=connection)


class EmptyQuerySet(QuerySet):
    def _do_query(self):
        return EmptyResultSet()


class ValueListResolver(object):
    """
    Class to resolve vertices in django objects
    """

    def __init__(self, model, fields, flat=False):
        """
        :param model: the current model
        :param fields: a cooked fields set (es_field, djfield, field)
        """
        self.fields = fields
        self.model = model
        self.flat = flat

    def __call__(self, *args, **kwargs):
        results = []
        from django.db import models

        es_fields = args[1].get("fields", {})
        for es_name, dj_name, djfield in self.fields:
            if es_name == "pk":
                value = args[1]["_id"]
            elif es_name == "id":
                value = es_fields.get(es_name, args[1]["_id"])
            else:
                value = es_fields.get(es_name, None)
            if value and isinstance(djfield, models.ForeignKey):
                djfield.rel.to.objects.get(pk=value)
            if self.flat:
                return value
            results.append(value)
        return tuple(results)


class ModelIteratorResolver(object):
    def __init__(self, model, items=None):
        self.model = model
        self.items = items
        self.iterpos = 0
        self._current_item = 0

    def __len__(self):
        return len(self.items)

    def count(self):
        return len(self.items)

    def total(self):
        return len(self.items)

    def __call__(self, *args, **kwargs):
        return list(self)

    def __getitem__(self, val):
        if not isinstance(val, (int, long, slice)):
            raise TypeError('%s indices must be integers, not %s' % (
                self.__class__.__name__, val.__class__.__name__))

        def get_start_end(val):
            if isinstance(val, slice):
                start = val.start
                if not start:
                    start = 0
                end = val.stop or len(self.items)
                if end < 0:
                    end = len(self.items) + end
                if end > len(self.items):
                    end = len(self.items)
                return start, end
            return val, val + 1

        start, end = get_start_end(val)

        if not isinstance(val, slice):
            if len(self.items) == 1:
                val = self.get_model(self.items[0])
                if val:
                    return val
                raise StopIteration
            raise IndexError
        return [v for v in [self.get_model(hit) for hit in self.items[start:end]] if v]

    def next(self):
        if len(self.items) == self.iterpos:
            raise StopIteration

        res = self.items[self.iterpos]
        val = self.get_model(res)
        if not val:
            return self.next()

        self.iterpos += 1
        if len(self.items) == self.iterpos:
            raise StopIteration
        return val

    def __iter__(self):
        self.iterpos = 0
        self._current_item = 0
        return self

    def get_model(self, object_id):
        try:
            return self.model.objects.get(pk=object_id)
        except self.model.DoesNotExist:
            self.items.remove(object_id)
            return None


class FacetHelper(dict):
    def __init__(self, model, facets):
        self._model = model
        self.update(facets)

    def __setattr__(self, key, value):
        if not self.__dict__.has_key(
                '_FacetHelper__initialised'):  # this test allows attributes to be set in the __init__ method
            return dict.__setattr__(self, key, value)
        elif self.__dict__.has_key(key):       # any normal attributes are handled normally
            dict.__setattr__(self, key, value)
        else:
            self.__setitem__(key, value)

    def get_as_list(self, name):
        terms = self.get(name, {}).get("terms", [])
        for term in terms:
            yield term["term"]
        raise StopIteration



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
from .exceptions import IndexAlreadyExistsException, IndexMissingException
from .utils import make_path
from .filters import Filter
from .mappings import Mapper
import six

class Indices(object):
    alias_params = ['filter', 'routing', 'search_routing', 'index_routing']

    def __init__(self, conn):
        self.conn = conn

    #TODO: clearcache segments templates

    #Alias Management - START
    def aliases(self, indices=None):
        """
        Retrieve the aliases of one or more indices.
        ( See :ref:`es-guide-reference-api-admin-indices-aliases`)

        :keyword indices: an index or a list of indices

        """
        path = self.conn._make_path(indices, (), '_aliases')
        return self.conn._send_request('GET', path)

    def get_alias(self, alias):
        """
        Get the index or indices pointed to by a given alias.
        (See :ref:`es-guide-reference-api-admin-indices-aliases`)

        :param alias: the name of an alias

        :return returns a list of index names.
        :raise IndexMissingException if the alias does not exist.

        """
        status = self.status([alias])
        return status['indices'].keys()

    def change_aliases(self, commands):
        """
        Change the aliases stored.
        (See :ref:`es-guide-reference-api-admin-indices-aliases`)

        :param commands: is a list of 3-tuples; (command, index, alias), where
                         `command` is one of "add" or "remove", and `index` and
                         `alias` are the index and alias to add or remove.

        """
        body = {'actions': [{command: dict(index=index, alias=alias, **params)}
                            for (command, index, alias, params) in commands]}
        return self.conn._send_request('POST', "_aliases", body)

    def _get_alias_params(self, **kwargs):
        ret = {}
        for name, value in kwargs.items():
            if name in self.alias_params and value:
                if isinstance(value, Filter):
                    ret[name] = value.serialize()
                else:
                    ret[name] = value

        return ret

    def add_alias(self, alias, indices, **kwargs):
        """
        Add an alias to point to a set of indices.
        (See :ref:`es-guide-reference-api-admin-indices-aliases`)

        :param alias: the name of an alias
        :param indices: a list of indices

        """
        indices = self.conn._validate_indices(indices)

        return self.change_aliases(['add', index, alias,
                                    self._get_alias_params(**kwargs)]
                                    for index in indices)

    def delete_alias(self, alias, indices):
        """
        Delete an alias.
        (See :ref:`es-guide-reference-api-admin-indices-aliases`)

        The specified index or indices are deleted from the alias, if they are
        in it to start with.  This won't report an error even if the indices
        aren't present in the alias.

        :param alias: the name of an alias
        :param indices: a list of indices
        """
        indices = self.conn._validate_indices(indices)

        return self.change_aliases(['remove', index, alias, {}] for index in indices)

    def set_alias(self, alias, indices, **kwargs):
        """
        Set an alias.
        (See :ref:`es-guide-reference-api-admin-indices-aliases`)

        This handles removing the old list of indices pointed to by the alias.

        Warning: there is a race condition in the implementation of this
        function - if another client modifies the indices which this alias
        points to during this call, the old value of the alias may not be
        correctly set.

        :param alias: the name of an alias
        :param indices: a list of indices
        """
        indices = self.conn._validate_indices(indices)
        try:
            old_indices = self.get_alias(alias)
        except IndexMissingException:
            old_indices = []
        commands = [['remove', index, alias, {}] for index in old_indices]
        commands.extend([['add', index, alias,
                          self._get_alias_params(**kwargs)] for index in indices])
        if len(commands) > 0:
            return self.change_aliases(commands)

    def stats(self, indices=None):
        """
        Retrieve the statistic of one or more indices
        (See :ref:`es-guide-reference-api-admin-indices-stats`)

        :keyword indices: an index or a list of indices
        """
        path = self.conn._make_path(indices, (), "_stats")
        return self.conn._send_request('GET', path)

    def status(self, indices=None):
        """
        Retrieve the status of one or more indices
        (See :ref:`es-guide-reference-api-admin-indices-status`)

        :keyword indices: an index or a list of indices
        """
        path = self.conn._make_path(indices, (), '_status', allow_all_indices=False)
        return self.conn._send_request('GET', path)
    # Following signature differs because field data isn't available at _all endpoint.   
    def field_stats(self, indices=''):
        """
        Retrieve the field data stats for one or more indices
        (See :ref:'es-guide-reference-api-admin-cluster-nodes-stats')

        :keyword indices: an index or a list of indices
        """
        path = self.conn._make_path(indices, (), '_stats','fielddata')
        return self.conn._send_request('GET', path)

    def create_index(self, index, settings=None):
        """
        Creates an index with optional settings.
        :ref:`es-guide-reference-api-admin-indices-create-index`

        :param index: the name of the index
        :keyword settings: a settings object or a dict containing settings

        """
        return self.conn._send_request('PUT', index, settings)

    def create_index_if_missing(self, index, settings=None):
        """Creates an index if it doesn't already exist.

        If supplied, settings must be a dictionary.

        :param index: the name of the index
        :keyword settings: a settings object or a dict containing settings
        """
        try:
            return self.create_index(index, settings)
        except IndexAlreadyExistsException as e:
            return e.result

    def delete_index(self, index):
        """
        Deletes an index.
        :ref:`es-guide-reference-api-admin-indices-delete-index`

        :param index: the name of the index

        """
        return self.conn._send_request('DELETE', index)

    def exists_index(self, index):
        """
        Check if an index exists.
        (See :ref:`es-guide-reference-api-admin-indices-indices-exists`)

        :param index: the name of the index
        """
        return self.conn._send_request('HEAD', index)

    def delete_index_if_exists(self, index):
        """
        Deletes an index if it exists.

        :param index: the name of the index

        """
        if self.exists_index(index):
            return self.delete_index(index)

    def get_indices(self, include_aliases=False):
        """
        Get a dict holding an entry for each index which exists.

        If include_alises is True, the dict will also contain entries for
        aliases.

        The key for each entry in the dict is the index or alias name.  The
        value is a dict holding the following properties:

         - num_docs: Number of documents in the index or alias.
         - alias_for: Only present for an alias: holds a list of indices which
           this is an alias for.

        """
        state = self.conn.cluster.state()
        status = self.status()
        result = {}
        indices_status = status['indices']
        indices_metadata = state['metadata']['indices']
        for index in sorted(indices_status.keys()):
            info = indices_status[index]
            try:
                num_docs = info['docs']['num_docs']
            except KeyError:
                num_docs = 0
            result[index] = dict(num_docs=num_docs)

            if not include_aliases:
                continue
            try:
                metadata = indices_metadata[index]
            except KeyError:
                continue
            for alias in metadata.get('aliases', []):
                try:
                    alias_obj = result[alias]
                except KeyError:
                    alias_obj = {}
                    result[alias] = alias_obj
                alias_obj['num_docs'] = alias_obj.get('num_docs', 0) + num_docs
                try:
                    alias_obj['alias_for'].append(index)
                except KeyError:
                    alias_obj['alias_for'] = [index]
        return result

    def get_closed_indices(self):
        """
        Get all closed indices.
        """
        state = self.conn.cluster.state()
        status = self.status()

        indices_metadata = set(state['metadata']['indices'].keys())
        indices_status = set(status['indices'].keys())

        return indices_metadata.difference(indices_status)

    def close_index(self, index):
        """
        Close an index.
        (See :ref:`es-guide-reference-api-admin-indices-open-close`)


        :param index: the name of the index

        """
        return self.conn._send_request('POST', "/%s/_close" % index)

    def open_index(self, index):
        """
        Open an index.
        (See :ref:`es-guide-reference-api-admin-indices-open-close`)

        :param index: the name of the index
        """
        return self.conn._send_request('POST', "/%s/_open" % index)

    def flush(self, indices=None, refresh=None):
        """
        Flushes one or more indices (clear memory)
        If a bulk is full, it sends it.

        (See :ref:`es-guide-reference-api-admin-indices-flush`)


        :keyword indices: an index or a list of indices
        :keyword refresh: set the refresh parameter

        """
        self.conn.force_bulk()
        path = self.conn._make_path(indices, '_flush')
        args = {}
        if refresh is not None:
            args['refresh'] = refresh
        return self.conn._send_request('POST', path, params=args)

    def refresh(self, indices=None, timesleep=None, timeout=0):
        """
        Refresh one or more indices
        If a bulk is full, it sends it.
        (See :ref:`es-guide-reference-api-admin-indices-refresh`)

        :keyword indices: an index or a list of indices
        :keyword timesleep: seconds to wait
        :keyword timeout: seconds to wait before timing out when waiting for
            the cluster's health.
        """
        self.conn.force_bulk()
        path = self.conn._make_path(indices, (), '_refresh', allow_all_indices=False)
        result = self.conn._send_request('POST', path)
        if timesleep:
            time.sleep(timesleep)
        self.conn.cluster.health(wait_for_status='green', timeout=timeout)
        return result

    def optimize(self, indices=None,
                 wait_for_merge=False,
                 max_num_segments=None,
                 only_expunge_deletes=False,
                 refresh=True,
                 flush=True):
        """
        Optimize one or more indices.
        (See :ref:`es-guide-reference-api-admin-indices-optimize`)


        :keyword indices: the list of indices to optimise.  If not supplied, all
                          default_indices are optimised.

        :keyword wait_for_merge: If True, the operation will not return until the merge has been completed.
                                 Defaults to False.

        :keyword max_num_segments: The number of segments to optimize to. To fully optimize the index, set it to 1.
                                   Defaults to half the number configured by the merge policy (which in turn defaults
                                   to 10).


        :keyword only_expunge_deletes: Should the optimize process only expunge segments with deletes in it.
                                       In Lucene, a document is not deleted from a segment, just marked as deleted.
                                       During a merge process of segments, a new segment is created that does have
                                       those deletes.
                                       This flag allow to only merge segments that have deletes. Defaults to false.

        :keyword refresh: Should a refresh be performed after the optimize. Defaults to true.

        :keyword flush: Should a flush be performed after the optimize. Defaults to true.

        """
        path = self.conn._make_path(indices, (), '_optimize')
        params = dict(
            wait_for_merge=wait_for_merge,
            only_expunge_deletes=only_expunge_deletes,
            refresh=refresh,
            flush=flush,
            )
        for k, v in params.items():
            params[k] = v and "true" or "false"
        if max_num_segments is not None:
            params['max_num_segments'] = max_num_segments
        return self.conn._send_request('POST', path, params=params)

    def analyze(self, text, index=None, analyzer=None, tokenizer=None, filters=None, field=None):
        """
        Performs the analysis process on a text and return the tokens breakdown of the text

        (See :ref:`es-guide-reference-api-admin-indices-optimize`)

        """
        if filters is None:
            filters = []
        argsets = 0
        args = {}

        if analyzer:
            args['analyzer'] = analyzer
            argsets += 1
        if tokenizer or filters:
            if tokenizer:
                args['tokenizer'] = tokenizer
            if filters:
                args['filters'] = ','.join(filters)
            argsets += 1
        if field:
            args['field'] = field
            argsets += 1

        if argsets > 1:
            raise ValueError('Argument conflict: Specify either analyzer, tokenizer/filters or field')

        if field and index is None:
            raise ValueError('field can only be specified with an index')

        path = make_path(index, '_analyze')
        return self.conn._send_request('POST', path, text, args)

    def gateway_snapshot(self, indices=None):
        """
        Gateway snapshot one or more indices
        (See :ref:`es-guide-reference-api-admin-indices-gateway-snapshot`)

        :keyword indices: a list of indices or None for default configured.
        """
        path = self.conn._make_path(indices, (), '_gateway', 'snapshot')
        return self.conn._send_request('POST', path)

    def put_mapping(self, doc_type=None, mapping=None, indices=None, ignore_conflicts=None):
        """
        Register specific mapping definition for a specific type against one or more indices.
        (See :ref:`es-guide-reference-api-admin-indices-put-mapping`)

        """
        if not isinstance(mapping, dict):
            if mapping is None:
                mapping = {}
            if hasattr(mapping, "as_dict"):
                mapping = mapping.as_dict()

        if doc_type:
            path = self.conn._make_path(indices, doc_type, "_mapping")
            if doc_type not in mapping:
                mapping = {doc_type: mapping}
        else:
            path = self.conn._make_path(indices, (), "_mapping")

        parameters = {}

        if ignore_conflicts is not None:
            parameters['ignore_conflicts'] = ignore_conflicts

        return self.conn._send_request('PUT', path, mapping, params=parameters)

    def get_mapping(self, doc_type=None, indices=None, raw=False):
        """
        Register specific mapping definition for a specific type against one or more indices.
        (See :ref:`es-guide-reference-api-admin-indices-get-mapping`)

        """
        if doc_type is None and indices is None:
            path = make_path("_mapping")
            is_mapping = False
        else:
            indices = self.conn._validate_indices(indices)
            if doc_type:
                path = make_path(','.join(indices), doc_type, "_mapping")
                is_mapping = True
            else:
                path = make_path(','.join(indices), "_mapping")
                is_mapping = False
        result = self.conn._send_request('GET', path)
        if raw:
            return result
        from pyes.mappings import Mapper
        mapper = Mapper(result, is_mapping=False,
                        connection=self.conn,
                        document_object_field=self.conn.document_object_field)
        if doc_type:
            return mapper.mappings[doc_type]
        return mapper

    def delete_mapping(self, index, doc_type):
        """
        Delete a typed JSON document type from a specific index.
        (See :ref:`es-guide-reference-api-admin-indices-delete-mapping`)

        """
        path = make_path(index, doc_type)
        return self.conn._send_request('DELETE', path)

    def get_settings(self, index=None):
        """
        Returns the current settings for an index.
        (See :ref:`es-guide-reference-api-admin-indices-get-settings`)

        """
        path = make_path(index, "_settings")
        return self.conn._send_request('GET', path)

    def update_settings(self, index, newvalues):
        """
        Update Settings of an index.
        (See  :ref:`es-guide-reference-api-admin-indices-update-settings`)

        """
        path = make_path(index, "_settings")
        return self.conn._send_request('PUT', path, newvalues)

class Cluster(object):
    def __init__(self, conn):
        self.conn = conn

    #TODO: node shutdown, update settings

    def shutdown(self, all_nodes=False, master=False, local=False, nodes=[],
                 delay=None):
        if all_nodes:
            path = make_path('_shutdown')
        elif master:
            path = make_path("_cluster", "nodes", "_master", "_shutdown")
        elif nodes:
            path = make_path("_cluster", "nodes", ",".join(nodes), "_shutdown")
        elif local:
            path = make_path("_cluster", "nodes", "_local", "_shutdown")
        if delay:
            try:
                int(delay)
                path += "?%s"%delay
            except ValueError:
                raise ValueError("%s is not a valid delay time"%delay)
            except TypeError:
                raise TypeError("%s is not of type int or string"%delay)
        if not path:
            # default action
            path = make_path('_shutdown')
        return self.conn._send_request('GET', path)



    def health(self, indices=None, level="cluster", wait_for_status=None,
               wait_for_relocating_shards=None, timeout=30):
        """
        Check the current :ref:`cluster health <es-guide-reference-api-admin-cluster-health>`.
        Request Parameters

        The cluster health API accepts the following request parameters:

        :param level: Can be one of cluster, indices or shards. Controls the
                        details level of the health information returned.
                        Defaults to *cluster*.
        :param wait_for_status: One of green, yellow or red. Will wait (until
                                the timeout provided) until the status of the
                                cluster changes to the one provided.
                                By default, will not wait for any status.
        :param wait_for_relocating_shards: A number controlling to how many
                                           relocating shards to wait for.
                                           Usually will be 0 to indicate to
                                           wait till all relocation have
                                           happened. Defaults to not to wait.
        :param timeout: A time based parameter controlling how long to wait
                        if one of the wait_for_XXX are provided.
                        Defaults to 30s.
        """
        if indices:
            path = make_path("_cluster", "health", ",".join(indices))
        else:
            path = make_path("_cluster", "health")
        mapping = {}
        if level != "cluster":
            if level not in ["cluster", "indices", "shards"]:
                raise ValueError("Invalid level: %s" % level)
            mapping['level'] = level
        if wait_for_status:
            if wait_for_status not in ["green", "yellow", "red"]:
                raise ValueError("Invalid wait_for_status: %s" % wait_for_status)
            mapping['wait_for_status'] = wait_for_status

            mapping['timeout'] = "%ds" % timeout
        return self.conn._send_request('GET', path, params=mapping)

    def state(self, filter_nodes=None, filter_routing_table=None,
                      filter_metadata=None, filter_blocks=None,
                      filter_indices=None):
        """
        Retrieve the :ref:`cluster state <es-guide-reference-api-admin-cluster-state>`.

        :param filter_nodes: set to **true** to filter out the **nodes** part
                             of the response.
        :param filter_routing_table: set to **true** to filter out the
                                     **routing_table** part of the response.
        :param filter_metadata: set to **true** to filter out the **metadata**
                                part of the response.
        :param filter_blocks: set to **true** to filter out the **blocks**
                              part of the response.
        :param filter_indices: when not filtering metadata, a comma separated
                               list of indices to include in the response.

        """
        path = make_path("_cluster", "state")
        parameters = {}

        if filter_nodes is not None:
            parameters['filter_nodes'] = filter_nodes

        if filter_routing_table is not None:
            parameters['filter_routing_table'] = filter_routing_table

        if filter_metadata is not None:
            parameters['filter_metadata'] = filter_metadata

        if filter_blocks is not None:
            parameters['filter_blocks'] = filter_blocks

        if filter_blocks is not None:
            if isinstance(filter_indices, six.string_types):
                parameters['filter_indices'] = filter_indices
            else:
                parameters['filter_indices'] = ",".join(filter_indices)

        return self.conn._send_request('GET', path, params=parameters)

    def nodes_info(self, nodes=None):
        """
        The cluster :ref:`nodes info <es-guide-reference-api-admin-cluster-state>` API allows to retrieve one or more (or all) of
        the cluster nodes information.
        """
        parts = ["_cluster", "nodes"]
        if nodes:
            parts.append(",".join(nodes))
        path = make_path(*parts)
        return self.conn._send_request('GET', path)


    def info(self):
        """
        The cluster :ref:`nodes info <es-guide-reference-api-admin-cluster-state>` API allows to retrieve one or more (or all) of
        the cluster nodes information.
        """
        return self.conn._send_request('GET', "/")


    def node_stats(self, nodes=None):
        """
        The cluster :ref:`nodes info <es-guide-reference-api-admin-cluster-nodes-stats>` API allows to retrieve one or more (or all) of
        the cluster nodes information.
        """
        parts = ["_cluster", "nodes", "stats"]
        if nodes:
            parts = ["_cluster", "nodes", ",".join(nodes), "stats"]

        path = make_path(*parts)
        return self.conn._send_request('GET', path)

    def node_field_stats(self, nodes=None):
        """
        Retrieve field stats per node(s) or cluster wide
        :ref:`nodes info <es-guide-reference-api-admin-cluster-nodes-stats>` API allows to retrieve one or more (or all) 
        the cluster nodes info about field data memory usage
        """
        parts = ['_nodes', 'stats', 'indices']
        if nodes:
            paths = ['_nodes', ",".join(nodes), 'stats', 'indices']

        path = make_path(*parts)
        return self.conn._send_request('GET', path)



# -*- coding: utf-8 -*-

class HighLighter(object):
    """
    This object manage the highlighting

    :arg pre_tags: list of tags before the highlighted text.
        importance is ordered..  ex. ``['<b>']``
    :arg post_tags: list of end tags after the highlighted text.
        should line up with pre_tags.  ex. ``['</b>']``
    :arg fields: list of fields to highlight
    :arg fragment_size: the size of the grament
    :arg number_or_fragments: the maximum number of fragments to
        return; if 0, then no fragments are returned and instead the
        entire field is returned and highlighted.
    :arg fragment_offset: controls the margin to highlight from

    Use this with a :py:class:`pyes.query.Search` like this::

        h = HighLighter(['<b>'], ['</b>'])
        s = Search(TermQuery('foo'), highlight=h)
    """

    def __init__(self, pre_tags=None, post_tags=None, fields=None, fragment_size=None, number_of_fragments=None,
                 fragment_offset=None, encoder=None):
        self.pre_tags = pre_tags
        self.post_tags = post_tags
        self.fields = fields or {}
        self.fragment_size = fragment_size
        self.number_of_fragments = number_of_fragments
        self.fragment_offset = fragment_offset
        self.encoder = encoder

    def add_field(self, name, fragment_size=150, number_of_fragments=3, fragment_offset=None, order="score", type=None):
        """
        Add a field to Highlinghter
        """
        data = {}
        if fragment_size:
            data['fragment_size'] = fragment_size
        if number_of_fragments is not None:
            data['number_of_fragments'] = number_of_fragments
        if fragment_offset is not None:
            data['fragment_offset'] = fragment_offset
        if type is not None:
            data['type'] = type
        data['order'] = order
        self.fields[name] = data

    def serialize(self):
        res = {}
        if self.pre_tags:
            res["pre_tags"] = self.pre_tags
        if self.post_tags:
            res["post_tags"] = self.post_tags
        if self.fragment_size:
            res["fragment_size"] = self.fragment_size
        if self.number_of_fragments:
            res["number_of_fragments"] = self.number_of_fragments
        if self.fragment_offset:
            res["fragment_offset"] = self.fragment_offset
        if self.encoder:
            res["encoder"] = self.encoder
        if self.fields:
            res["fields"] = self.fields
        else:
            res["fields"] = {"_all": {}}
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from future import print_function

from pyes.es import ES
from pyes import mappings

def mappings_to_code(obj, doc_count=0):
    result = []
    odict = obj.as_dict()
    if isinstance(obj, (mappings.DocumentObjectField, mappings.ObjectField, mappings.NestedObject)):
        properties = odict.pop("properties", [])
        doc_count += 1
        kwargs = ["name=%r" % obj.name,
                  "type=%r" % odict.pop("type")] +\
                 ["%s=%r" % (k, odict[k]) for k in sorted(odict.keys())]
        result.append(
            "doc%d=" % doc_count + str(type(obj)).split(".")[-1].strip("'>") + "(" + ', '.join(kwargs) + ")")
        for k in sorted(obj.properties.keys()):
            result.extend(mappings_to_code(obj.properties[k], doc_count))
    else:
        kwargs = ["name=%r" % obj.name,
                  "type=%r" % odict.pop("type"),
                  "store=%r" % obj.store,
                  "index=%r" % odict.pop("index")] +\
                 ["%s=%r" % (k, odict[k]) for k in sorted(odict.keys())]
        result.append("doc%d.add_property(" % doc_count +\
                      str(type(obj)).split(".")[-1].strip("'>") + "(" +\
                      ', '.join(kwargs) + "))")

    return result

if __name__ == '__main__':
    es = ES("192.168.1.1:9200")
    res = mappings_to_code(es.mappings.get_doctype("twitter", "twitter"))
    print("\n".join(res))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
import codecs
import platform

try:
    from setuptools import setup, find_packages, Command
    from setuptools.command.test import test as TestCommand
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages, Command
    from setuptools.command.test import test as TestCommand

# Extract distribution meta values, hint taken from Celery <http://celeryproject.org>

import re
re_meta = re.compile(r'__(\w+?)__\s*=\s*(.*)')
re_vers = re.compile(r'VERSION\s*=\s*\((.*?)\)')
re_doc = re.compile(r'^"""(.+?)"""')
rq = lambda s: s.strip("\"'")

def add_default(m):
    attr_name, attr_value = m.groups()
    return ((attr_name, rq(attr_value)), )


def add_version(m):
    v = list(map(rq, m.groups()[0].split(", ")))
    return (("VERSION", ".".join(v[0:3]) + "".join(v[3:])), )


def add_doc(m):
    return (("doc", m.groups()[0]), )

pats = {re_meta: add_default,
        re_vers: add_version,
        re_doc: add_doc}
here = os.path.abspath(os.path.dirname(__file__))
meta_fh = open(os.path.join(here, "pyes/__init__.py"))
try:
    meta = {}
    for line in meta_fh:
        for pattern, handler in pats.items():
            m = pattern.match(line.strip())
            if m:
                meta.update(handler(m))
finally:
    meta_fh.close()


class QuickRunTests(TestCommand):
    extra_env = dict(SKIP_RLIMITS=1, QUICKTEST=1)

    def run(self, *args, **kwargs):
        for env_name, env_value in self.extra_env.items():
            os.environ[env_name] = str(env_value)
        TestCommand.run(self, *args, **kwargs)


install_requires = ["urllib3>=1.7", "six>=1.5.2"]

#if not sys.platform.startswith("java"):
#    install_requires += [ "thrift", ]    
try:
    import importlib
except ImportError:
    install_requires.append("importlib")

try:
    # For Python >= 2.6
    import json
except ImportError:
    # For Python < 2.6 or people using a newer version of simplejson
    install_requires.append("simplejson")

try:
    from collections import OrderedDict
except ImportError:
    # Python 2.6 or earlier, use backport
    #from ordereddict import OrderedDict
    install_requires.append("ordereddict")

py_version = sys.version_info
if not sys.platform.startswith("java") and sys.version_info < (2, 6):
    install_requires.append("multiprocessing==2.6.2.1")

if os.path.exists("README.rst"):
    long_description = codecs.open("README.rst", "r", "utf-8").read()
else:
    long_description = "See http://pypi.python.org/pypi/pyes"

setup(
    name='pyes',
    version=meta['VERSION'],
    description="Python Elastic Search driver",
    author=meta['author'],
    author_email=meta['contact'],
    url=meta['homepage'],
    platforms=["any"],
    license="BSD",
    packages=find_packages(exclude=['ez_setup', 'tests', 'tests.*', "docs.*"]),
    scripts=[],
    zip_safe=False,
    install_requires=install_requires,
    tests_require=['nose', 'nose-cover3', 'unittest2', 'simplejson'],
    cmdclass={"quicktest": QuickRunTests},
    test_suite="nose.collector",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search'
    ],
    entry_points={
        'console_scripts': [],
    },
    long_description=long_description,
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from types import NoneType
import datetime
from django.db import models
import uuid

__author__ = 'Alberto Paro'
__all__ = ["get_values"]


#--- taken from http://djangosnippets.org/snippets/2278/

def get_values(instance, go_into={}, exclude=(), extra=(), skip_none=False):
    """
    Transforms a django model instance into an object that can be used for
    serialization.
    @param instance(django.db.models.Model) - the model in question
    @param go_into(dict) - relations with other models that need expanding
    @param exclude(tuple) - fields that will be ignored
    @param extra(tuple) - additional functions/properties which are not fields
    @param skip_none(bool) - skip None field

    Usage:
    get_values(MyModel.objects.get(pk=187),
               {'user': {'go_into': ('clan',),
                         'exclude': ('crest_blob',),
                         'extra': ('get_crest_path',)}},
               ('image'))

    """
    from django.db.models.manager import Manager
    from django.db.models import Model

    SIMPLE_TYPES = (int, long, str, list, dict, tuple, bool, float, bool,
                    unicode, NoneType)

    if not isinstance(instance, Model):
        raise TypeError("Argument is not a Model")

    value = {
        'pk': instance.pk,
        }

    # check for simple string instead of tuples
    # and dicts; this is shorthand syntax
    if isinstance(go_into, str):
        go_into = {go_into: {}}

    if isinstance(exclude, str):
        exclude = (exclude,)

    if isinstance(extra, str):
        extra = (extra,)

    # process the extra properties/function/whatever
    for field in extra:
        property = getattr(instance, field)

        if callable(property):
            property = property()

        if skip_none and property is None:
            continue
        elif isinstance(property, SIMPLE_TYPES):
            value[field] = property
        else:
            value[field] = repr(property)

    field_options = instance._meta.get_all_field_names()
    for field in field_options:
        try:
            property = getattr(instance, field)
        except:
            continue
        if skip_none and property is None:
            continue

        if field in exclude or field[0] == '_' or isinstance(property, Manager):
            # if it's in the exclude tuple, ignore it
            # if it's a "private" field, ignore it
            # if it's an instance of manager (this means a more complicated
            # relationship), ignore it
            continue
        elif go_into.has_key(field):
            # if it's in the go_into dict, make a recursive call for that field
            try:
                field_go_into = go_into[field].get('go_into', {})
            except AttributeError:
                field_go_into = {}

            try:
                field_exclude = go_into[field].get('exclude', ())
            except AttributeError:
                field_exclude = ()

            try:
                field_extra = go_into[field].get('extra', ())
            except AttributeError:
                field_extra = ()

            value[field] = get_values(property,
                field_go_into,
                field_exclude,
                field_extra, skip_none=skip_none)
        else:
            if isinstance(property, Model):
                # if it's a model, we need it's PK #
                value[field] = property.pk
            elif isinstance(property, (datetime.date,
                                       datetime.time,
                                       datetime.datetime)):
                value[field] = property
            else:
                # else, we just put the value #
                if callable(property):
                    property = property()

                if isinstance(property, SIMPLE_TYPES):
                    value[field] = property
                else:
                    value[field] = repr(property)

    return value


class EmbeddedModel(models.Model):
    _embedded_in = None

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if self.pk is None:
            self.pk = str(uuid.uuid4())
        if self._embedded_in  is None:
            raise RuntimeError("Invalid save")
        self._embedded_in.save()

    def serialize(self):
        if self.pk is None:
            self.pk = "TODO"
            self.id = self.pk
        result = {'_app': self._meta.app_label,
                  '_model': self._meta.model_name,
                  '_id': self.pk}
        for field in self._meta.fields:
            result[field.attname] = getattr(self, field.attname)
        return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import logging
import unittest
from pprint import pprint
from pyes.es import ES
from pyes.helpers import SettingsBuilder

"""
Unit tests for pyes.  These require an es server with thrift plugin running on the default port (localhost:9500).
"""

def get_conn(*args, **kwargs):
    return ES(("http", "127.0.0.1", 9200), *args, **kwargs)


class ESTestCase(unittest.TestCase):
    def setUp(self):
        self.log = open("/tmp/%s.sh"%self._testMethodName, "wb")
        self.conn = get_conn(timeout=300.0, log_curl=True, dump_curl=self.log)#incremented timeout for debugging
        self.index_name = "test-index"
        self.document_type = "test-type"
        self.conn.indices.delete_index_if_exists(self.index_name)

    def tearDown(self):
        self.conn.indices.delete_index_if_exists(self.index_name)
        if self.log:
            self.log.close()

    def assertResultContains(self, result, expected):
        for (key, value) in expected.items():
            found = False
            try:
                found = value == result[key]
            except KeyError:
                if result.has_key('meta'):
                    found = value == result['meta'][key]
            self.assertEqual(True, found)

    def checkRaises(self, excClass, callableObj, *args, **kwargs):
        """Assert that calling callableObj with *args and **kwargs raises an
        exception of type excClass, and return the exception object so that
        further tests on it can be performed.
        """
        try:
            callableObj(*args, **kwargs)
        except excClass as e:
            return e
        else:
            raise self.failureException("Expected exception %s not raised" % excClass)

    def get_datafile(self, filename):
        """
        Returns a the content of a test file
        """
        return open(os.path.join(os.path.dirname(__file__), "data", filename), "rb").read()

    def get_datafile_path(self, filename):
        """
        Returns a the content of a test file
        """
        return os.path.join(os.path.dirname(__file__), "data", filename)

    def dump(self, result):
        """
        dump to stdout the result
        """
        pprint(result)

    def init_default_index(self):
        settings = SettingsBuilder({'index.number_of_replicas': 0,
                                     "index.number_of_shards": 1})
        from pyes.mappings import DocumentObjectField
        from pyes.mappings import IntegerField
        from pyes.mappings import NestedObject
        from pyes.mappings import StringField, DateField

        docmapping = DocumentObjectField(name=self.document_type)
        docmapping.add_property(
            StringField(name="parsedtext", store=True, term_vector="with_positions_offsets", index="analyzed"))
        docmapping.add_property(
            StringField(name="name", store=True, term_vector="with_positions_offsets", index="analyzed"))
        docmapping.add_property(
            StringField(name="title", store=True, term_vector="with_positions_offsets", index="analyzed"))
        docmapping.add_property(IntegerField(name="position", store=True))
        docmapping.add_property(DateField(name="date", store=True))
        docmapping.add_property(StringField(name="uuid", store=True, index="not_analyzed"))
        nested_object = NestedObject(name="nested")
        nested_object.add_property(StringField(name="name", store=True))
        nested_object.add_property(StringField(name="value", store=True))
        nested_object.add_property(IntegerField(name="num", store=True))
        docmapping.add_property(nested_object)
        settings.add_mapping(docmapping)

        self.conn.ensure_index(self.index_name, settings)


def setUp():
    """Package level setup.

    For tests which don't modify the index, we don't want to have the overhead
    of setting up a test index, so we just set up test-pindex once, and use it
    for all tests.

    """
    mapping = {
        u'parsedtext': {
            'boost': 1.0,
            'index': 'analyzed',
            'store': 'yes',
            'type': u'string',
            "term_vector": "with_positions_offsets"},
        u'name': {
            'boost': 1.0,
            'index': 'analyzed',
            'store': 'yes',
            'type': u'string',
            "term_vector": "with_positions_offsets"},
        u'title': {
            'boost': 1.0,
            'index': 'analyzed',
            'store': 'yes',
            'type': u'string',
            "term_vector": "with_positions_offsets"},
        u'pos': {
            'store': 'yes',
            'type': u'integer'},
        u'doubles': {
            'store': 'yes',
            'type': u'double'},
        u'uuid': {
            'boost': 1.0,
            'index': 'not_analyzed',
            'store': 'yes',
            'type': u'string'}}

    conn = get_conn(log_curl=True)
    conn.indices.delete_index_if_exists("test-pindex")
    conn.indices.create_index("test-pindex")
    conn.indices.put_mapping("test-type", {'properties': mapping}, ["test-pindex"])
    conn.index({"name": "Joe Tester", "parsedtext": "Joe Testere nice guy", "uuid": "11111", "position": 1,
                "doubles": [1.0, 2.0, 3.0]}, "test-pindex", "test-type", 1)
    conn.index({"name": "Bill Baloney", "parsedtext": "Joe Testere nice guy", "uuid": "22222", "position": 2,
                "doubles": [0.1, 0.2, 0.3]}, "test-pindex", "test-type", 2)
    conn.indices.refresh(["test-pindex"])


def tearDown():
    """Remove the package level index.

    """
    conn = get_conn()
    conn.indices.delete_index_if_exists("test-pindex")
